@@ -230,12 +230,224 @@ struct ZoneWK4
 	shared_ptr<WPSStream> m_stream;
 };
 
+//! the graphic zone of a LotusGraph for 123 pc
+struct ZonePc
+{
+	//! the different type
+	enum Type { Arc, Chart, Ellipse, FreeHand, Line, Picture, Polygon, Rect, Set, TextBox, Unknown };
+	//! constructor
+	explicit ZonePc(shared_ptr<WPSStream> stream) :
+		m_type(Unknown), m_isGroup(false), m_groupLastPosition(0), m_numPoints(0), m_vertices(), m_isRoundRect(false), m_stream(stream), m_box(), m_translate(0,0), m_rotate(0), m_arrows(0), m_textBoxEntry(), m_pictureData(), m_pictureHeaderRead(0),
+		m_isSent(false), m_extra("")
+	{
+		for (int i=0; i<2; ++i) m_graphicId[i]=-1;
+	}
+	//! returns a graphic shape corresponding to the main form (and the origin)
+	bool getGraphicShape(WPSGraphicShape &shape, WPSPosition &pos) const;
+	//! returns a transformation corresponding to the shape
+	WPSTransformation getTransformation() const
+	{
+		WPSTransformation res;
+		if (m_rotate<0||m_rotate>0)
+			res = WPSTransformation::rotation(-m_rotate, m_box.center());
+		if (m_translate!=Vec2f(0,0))
+			res = WPSTransformation::translation(m_translate) * res;
+		return res;
+	}
+	//! operator<<
+	friend std::ostream &operator<<(std::ostream &o, ZonePc const &z)
+	{
+		switch (z.m_type)
+		{
+		case Arc:
+			o << "arc,";
+			break;
+		case Chart:
+			o << "chart,";
+			break;
+		case Ellipse:
+			o << "ellipse,";
+			break;
+		case FreeHand:
+			o << "freeHand,";
+			if (z.m_numPoints) o << "N=" << z.m_numPoints << ",";
+			break;
+		case Line:
+			o << "line,";
+			if (z.m_numPoints!=2) o << "N=" << z.m_numPoints << ",";
+			break;
+		case Picture:
+			o << "picture,";
+			break;
+		case Polygon:
+			o << "polygon,";
+			if (z.m_numPoints) o << "N=" << z.m_numPoints << ",";
+			break;
+		case Rect:
+			if (z.m_isRoundRect)
+				o << "rect[round],";
+			else
+				o << "rect,";
+			break;
+		case Set:
+			if (z.m_isGroup)
+				o << "group,";
+			else
+				o << "set,";
+			break;
+		case TextBox:
+			o << "textBox,";
+			break;
+		case Unknown:
+		default:
+			break;
+		}
+		o << "dim=" << z.m_box << ",";
+		if (z.m_translate!=Vec2f(0,0))
+			o << "translation=" << z.m_translate << ",";
+		if (z.m_rotate<0||z.m_rotate>0)
+			o << "rotation=" << z.m_rotate << ",";
+		for (int i=0; i<2; ++i)
+		{
+			if (z.m_graphicId[i]<0) continue;
+			o << (i==0 ? "style" : "shadow") << "=GS" << z.m_graphicId[i] << ",";
+		}
+		if (z.m_arrows&1)
+			o << "arrows[beg],";
+		if (z.m_arrows&2)
+			o << "arrows[end],";
+		o << z.m_extra << ",";
+		return o;
+	}
+	//! the zone type
+	Type m_type;
+	//! true if the set is a group
+	bool m_isGroup;
+	//! the group last position
+	size_t m_groupLastPosition;
+	//! the number of points of a line or a polygon
+	int m_numPoints;
+	//! the list of points for a polygon
+	std::vector<Vec2f> m_vertices;
+	//! true if the rect is a round rect
+	bool m_isRoundRect;
+	//! the stream
+	shared_ptr<WPSStream> m_stream;
+	//! the bdbox
+	WPSBox2f m_box;
+	//! the translation
+	Vec2f m_translate;
+	//! the rotation
+	float m_rotate;
+	//! the graphic style id and the shadow style
+	int m_graphicId[2];
+	//! the line arrows
+	int m_arrows;
+	//! the text box entry
+	WPSEntry m_textBoxEntry;
+	//! the picture data
+	librevenge::RVNGBinaryData m_pictureData;
+	//! the number of read byte of the header
+	int m_pictureHeaderRead;
+	//! a flag to know if the zone has already be sent
+	mutable bool m_isSent;
+	//! extra data
+	std::string m_extra;
+};
+
+bool ZonePc::getGraphicShape(WPSGraphicShape &shape, WPSPosition &pos) const
+{
+	pos=WPSPosition(m_box[0],m_box.size(), librevenge::RVNG_POINT);
+	pos.setRelativePosition(WPSPosition::Page);
+	WPSBox2f box(Vec2f(0,0), m_box.size());
+	switch (m_type)
+	{
+	case Line:
+	{
+		// we need to recompute the bdbox
+		float bounds[4]= {m_box[0][0],m_box[0][1],m_box[1][0],m_box[1][1]};
+		for (int i=0; i<2; ++i)
+		{
+			if (bounds[i]<=bounds[i+2]) continue;
+			bounds[i]=bounds[i+2];
+			bounds[i+2]=m_box[0][i];
+		}
+		WPSBox2f realBox(Vec2f(bounds[0],bounds[1]),Vec2f(bounds[2],bounds[3]));
+		pos=WPSPosition(realBox[0],realBox.size(), librevenge::RVNG_POINT);
+		pos.setRelativePosition(WPSPosition::Page);
+		shape=WPSGraphicShape::line(m_box[0]-realBox[0], m_box[1]-realBox[0]);
+		return true;
+	}
+	case Ellipse:
+		shape=WPSGraphicShape::circle(box);
+		return true;
+	case Rect:
+		if (m_isRoundRect)
+			shape=WPSGraphicShape::rectangle(box, Vec2f(5,5));
+		else
+			shape=WPSGraphicShape::rectangle(box);
+		return true;
+	case Arc: // checkme only work if no flip
+		shape=WPSGraphicShape::arc(box, WPSBox2f(Vec2f(-box[1][0],0),Vec2f(box[1][0],2*box[1][1])), Vec2f(0,90));
+		return true;
+	case FreeHand:
+	case Polygon:
+		if (m_vertices.empty())
+		{
+			WPS_DEBUG_MSG(("LotusGraphInternal::ZonePc::getGraphicShape: sorry, can not find the polygon vertices\n"));
+			return false;
+		}
+		shape=WPSGraphicShape::polygon(box);
+		shape.m_vertices=m_vertices;
+		shape.m_vertices.push_back(m_vertices[0]);
+		return true;
+	case Set:
+	case TextBox:
+		return true;
+	case Chart:
+	case Picture:
+	case Unknown:
+	default:
+		break;
+	}
+	static bool first=true;
+	if (first)
+	{
+		first=false;
+		WPS_DEBUG_MSG(("LotusGraphInternal::ZonePc::getGraphicShape: sorry, sending some graph types is not implemented\n"));
+	}
+	return false;
+}
+//! a list of ZonePc of a LotusGraph for 123 pc
+struct ZonePcList
+{
+	//! constructor
+	ZonePcList() : m_zones(), m_groupBeginStack()
+	{
+	}
+	//! returns true if the number of graphic zone is empty
+	bool empty() const
+	{
+		for (size_t i=0; i<m_zones.size(); ++i)
+		{
+			if (m_zones[i])
+				return false;
+		}
+		return true;
+	}
+	//! the list of zones
+	std::vector<shared_ptr<ZonePc> > m_zones;
+	//! the stack indicating the beginning of each group
+	std::stack<size_t> m_groupBeginStack;
+};
+
 //! the state of LotusGraph
 struct State
 {
 	//! constructor
 	State() :  m_version(-1), m_actualSheetId(-1),
-		m_sheetIdZoneMacMap(), m_actualZoneMac(), m_sheetIdZoneWK4Map(), m_actualZoneWK4()
+		m_sheetIdZoneMacMap(), m_actualZoneMac(), m_sheetIdZoneWK4Map(), m_actualZoneWK4(),
+		m_sheetIdZonePcListMap(), m_actualZonePc(), m_zIdToSheetIdMap()
 	{
 	}
 
@@ -251,6 +463,12 @@ struct State
 	std::multimap<int, shared_ptr<ZoneWK4> > m_sheetIdZoneWK4Map;
 	//! a pointer to the actual zone
 	shared_ptr<ZoneWK4> m_actualZoneWK4;
+	//! a map sheetid to zone
+	std::map<int, ZonePcList> m_sheetIdZonePcListMap;
+	//! a pointer to the actual zone
+	shared_ptr<ZonePc> m_actualZonePc;
+	//! a map sheet zone id to final sheet id map
+	std::map<int,int> m_zIdToSheetIdMap;
 };
 
 //! Internal: the subdocument of a LotusGraphc
@@ -329,6 +547,11 @@ void LotusGraph::cleanState()
 	m_state.reset(new LotusGraphInternal::State);
 }
 
+void LotusGraph::updateState(std::map<int,int> const &zIdToSheetIdMap)
+{
+	m_state->m_zIdToSheetIdMap=zIdToSheetIdMap;
+}
+
 int LotusGraph::version() const
 {
 	if (m_state->m_version<0)
@@ -338,8 +561,17 @@ int LotusGraph::version() const
 
 bool LotusGraph::hasGraphics(int sheetId) const
 {
-	return m_state->m_sheetIdZoneMacMap.find(sheetId)!=m_state->m_sheetIdZoneMacMap.end() ||
-	       m_state->m_sheetIdZoneWK4Map.find(sheetId)!=m_state->m_sheetIdZoneWK4Map.end() ;
+	if (m_state->m_sheetIdZoneMacMap.find(sheetId)!=m_state->m_sheetIdZoneMacMap.end() ||
+	        m_state->m_sheetIdZoneWK4Map.find(sheetId)!=m_state->m_sheetIdZoneWK4Map.end())
+		return true;
+	if (m_state->m_zIdToSheetIdMap.find(sheetId)!=m_state->m_zIdToSheetIdMap.end())
+	{
+		int finalId=m_state->m_zIdToSheetIdMap.find(sheetId)->second;
+		if (m_state->m_sheetIdZonePcListMap.find(finalId)!=m_state->m_sheetIdZonePcListMap.end()
+		        && !m_state->m_sheetIdZonePcListMap.find(finalId)->second.empty())
+			return true;
+	}
+	return false;
 }
 
 ////////////////////////////////////////////////////////////
@@ -1005,6 +1237,17 @@ void LotusGraph::sendGraphics(int sheetId)
 			}
 		}
 	}
+	if (m_state->m_zIdToSheetIdMap.find(sheetId)!=m_state->m_zIdToSheetIdMap.end())
+	{
+		int finalId=m_state->m_zIdToSheetIdMap.find(sheetId)->second;
+		if (m_state->m_sheetIdZonePcListMap.find(finalId)!=m_state->m_sheetIdZonePcListMap.end())
+		{
+			LotusGraphInternal::ZonePcList const &zoneList=m_state->m_sheetIdZonePcListMap.find(finalId)->second;
+			WPSTransformation transform;
+			for (size_t i=0; i<zoneList.m_zones.size(); ++i)
+				sendZone(zoneList, i, transform);
+		}
+	}
 }
 
 bool LotusGraph::readZoneBeginC9(shared_ptr<WPSStream> stream)
@@ -1296,7 +1539,9 @@ bool LotusGraph::readGraphic(shared_ptr<WPSStream> stream)
 		WPSBox2f box;
 		for (size_t i=0; i<3; ++i)
 		{
-			vertices[i]=Vec2f(float(libwps::readU16(input)),float(libwps::readU16(input)));
+			int pts[2];
+			for (int j=0; j<2; ++j) pts[j]=int(libwps::readU16(input));
+			vertices[i]=Vec2f(float(pts[0]),float(pts[1]));
 			if (i==0) box=WPSBox2f(vertices[i],vertices[i]);
 			else box=box.getUnion(WPSBox2f(vertices[i],vertices[i]));
 			f << vertices[i] << ",";
@@ -1321,7 +1566,9 @@ bool LotusGraph::readGraphic(shared_ptr<WPSStream> stream)
 		WPSBox2f box;
 		for (size_t i=0; i<size_t(N); ++i)
 		{
-			vertices[i]=Vec2f(float(libwps::readU16(input)),float(libwps::readU16(input)));
+			int pts[2];
+			for (int j=0; j<2; ++j) pts[j]=int(libwps::readU16(input));
+			vertices[i]=Vec2f(float(pts[0]),float(pts[1]));
 			if (i==0) box=WPSBox2f(vertices[i],vertices[i]);
 			else box=box.getUnion(WPSBox2f(vertices[i],vertices[i]));
 			f << vertices[i] << ",";
@@ -1369,7 +1616,9 @@ bool LotusGraph::readGraphic(shared_ptr<WPSStream> stream)
 		f << "pts=[";
 		for (int i=0; i<4; ++i)
 		{
-			Vec2f pt(float(libwps::readU16(input)), float(libwps::readU16(input)));
+			int pts[2];
+			for (int j=0; j<2; ++j) pts[j]=int(libwps::readU16(input));
+			Vec2f pt=Vec2f(float(pts[0]), float(pts[1]));
 			f << pt << ",";
 			if (i==0)
 				box=WPSBox2f(pt,pt);
@@ -1389,9 +1638,9 @@ bool LotusGraph::readGraphic(shared_ptr<WPSStream> stream)
 		f << "dim=";
 		for (int i=0; i<2; ++i)
 		{
-			f << int(libwps::readU16(input)) << "x" << int(libwps::readU16(input));
-			if (i==0) f << "<->";
-			else f << ",";
+			int pts[2];
+			for (int j=0; j<2; ++j) pts[j]=int(libwps::readU16(input));
+			f << Vec2i(pts[0],pts[1]) << (i==0 ? "<->" : ",");
 		}
 		break;
 	case 10: // group, useme
@@ -1399,7 +1648,11 @@ bool LotusGraph::readGraphic(shared_ptr<WPSStream> stream)
 			break;
 		f << "pts=[";
 		for (int i=0; i<2; ++i)
-			f << int(libwps::readU16(input)) << "x" << int(libwps::readU16(input)) << ",";
+		{
+			int pts[2];
+			for (int j=0; j<2; ++j) pts[j]=int(libwps::readU16(input));
+			f << Vec2i(pts[0],pts[1]) << ",";
+		}
 		f << "],";
 		val=int(libwps::readU16(input));
 		if (val!=1) f << "g0=" << val << ",";
@@ -1437,7 +1690,12 @@ bool LotusGraph::readGraphic(shared_ptr<WPSStream> stream)
 	{
 		if (sz!=0x43) break;
 		Vec2i dim[2];
-		for (int i=0; i<2; ++i) dim[i]=Vec2i(int(libwps::readU16(input)), int(libwps::readU16(input)));
+		for (int i=0; i<2; ++i)
+		{
+			int pts[2];
+			for (int j=0; j<2; ++j) pts[j]=int(libwps::readU16(input));
+			dim[i]=Vec2i(pts[0],pts[1]);
+		}
 		zone->m_pictureDim=WPSBox2i(dim[0], dim[1]);
 		f << "dim=" << dim << ",";
 		val=int(libwps::readU16(input)); // big number 2590..3262
@@ -1508,8 +1766,9 @@ bool LotusGraph::readFrame(shared_ptr<WPSStream> stream)
 	{
 		int row=int(libwps::readU16(input));
 		int col=int(libwps::readU8(input));
-		// first in 100.*char, second in point
-		Vec2f decal(8.f*float(libwps::readU16(input))/100.f, float(libwps::readU16(input)));
+		int pts[2]; // first in 100.*char, second in point
+		for (int i=0; i<2; ++i) pts[i]=int(libwps::readU16(input));
+		Vec2f decal(8.f*float(pts[0])/100.f, float(pts[1]));
 		Vec2i cell(col, row);
 		if (c==0 && zone)
 		{
@@ -1671,4 +1930,611 @@ void LotusGraph::sendTextBoxWK4(shared_ptr<WPSStream> stream, WPSEntry const &en
 	}
 }
 
+void LotusGraph::sendZone(LotusGraphInternal::ZonePcList const &zoneList, size_t id, WPSTransformation &transf)
+{
+	if (!m_listener)
+	{
+		WPS_DEBUG_MSG(("LotusGraph::sendZone: can not find the listener\n"));
+		return;
+	}
+	std::vector<shared_ptr<LotusGraphInternal::ZonePc> > const &zones=zoneList.m_zones;
+	if (id>=zones.size())
+	{
+		WPS_DEBUG_MSG(("LotusGraph::sendZone: can not find the sub zone %d\n", int(id)));
+		return;
+	}
+	shared_ptr<LotusGraphInternal::ZonePc> zone=zones[id];
+	if (!zone || zone->m_isSent) return;
+	zone->m_isSent=true;
+	WPSTransformation finalTrans=transf*zone->getTransformation();
+	if (zone->m_type==zone->Set)
+	{
+		if (!zone->m_isGroup || zone->m_groupLastPosition+1<=id) return;
+		WPSPosition pos(zone->m_box[0],zone->m_box.size(), librevenge::RVNG_POINT);
+		pos.setRelativePosition(WPSPosition::Page);
+		if (!m_listener->openGroup(pos)) return;
+		for (size_t i=id+1; i<zone->m_groupLastPosition; ++i)
+			sendZone(zoneList, i, finalTrans);
+		m_listener->closeGroup();
+		return;
+	}
+	if (zone->m_type==zone->Picture)
+	{
+		if (!zone->m_pictureData.empty())
+		{
+			WPSPosition pos(zone->m_box[0],zone->m_box.size(), librevenge::RVNG_POINT);
+			pos.setRelativePosition(WPSPosition::Page);
+			m_listener->insertPicture(pos, zone->m_pictureData);
+#ifdef DEBUG_WITH_FILES
+			std::stringstream s;
+			static int fileId=0;
+			s << "Pict" << ++fileId << ".emf";
+			libwps::Debug::dumpFile(zone->m_pictureData, s.str().c_str());
+#endif
+		}
+		return;
+	}
+
+	WPSGraphicShape shape;
+	WPSPosition pos;
+	if (!zone->getGraphicShape(shape, pos))
+		return;
+	WPSGraphicStyle style;
+	if (zone->m_graphicId[0]>=0)
+		m_styleManager->updateGraphicStyle(zone->m_graphicId[0], style);
+	if (zone->m_type==zone->TextBox)
+	{
+		shared_ptr<LotusGraphInternal::SubDocument> doc
+		(new LotusGraphInternal::SubDocument(zone->m_stream, *this, zone->m_textBoxEntry, 1));
+		m_listener->insertTextBox(pos, doc, style);
+		return;
+	}
+	if (zone->m_type==zone->Line)
+	{
+		if (zone->m_arrows&1)
+			style.m_arrows[1]=true;
+		if (zone->m_arrows&2)
+			style.m_arrows[0]=true;
+	}
+	if (finalTrans.isIdentity())
+		m_listener->insertPicture(pos, shape, style);
+	else
+	{
+		// checkme: ok for translation but not for rotation...
+		WPSGraphicShape finalShape=shape.transform(finalTrans);
+		pos.setOrigin(finalTrans*pos.origin());
+		pos.setSize(finalShape.getBdBox().size());
+		m_listener->insertPicture(pos, finalShape, style);
+	}
+}
+
+bool LotusGraph::readGraphZone(shared_ptr<WPSStream> stream, int zId)
+{
+	if (!stream) return false;
+	RVNGInputStreamPtr &input = stream->m_input;
+	libwps::DebugFile &ascFile=stream->m_ascii;
+	libwps::DebugStream f;
+	float const unit=version()>=5 ? 1.f/16.f : 1.f/256.f;
+	long pos = input->tell();
+	int id = (int) libwps::readU8(input);
+	if (libwps::readU8(input) != 3)
+	{
+		input->seek(pos, librevenge::RVNG_SEEK_SET);
+		return false;
+	}
+	long sz = (long) libwps::readU16(input);
+	long endPos=pos+4+sz;
+	if (sz<0 || !stream->checkFilePosition(endPos))
+	{
+		input->seek(pos, librevenge::RVNG_SEEK_SET);
+		return false;
+	}
+	if (zId)
+		f << "Entries(GraphZone)[Z" << zId << "]:";
+	else
+		f << "Entries(GraphZone)[_]:";
+	if (id<0x80)
+		m_state->m_actualZonePc.reset();
+	int val;
+	switch (id)
+	{
+	case 0: // rare, when it exists, present in sheet zone
+		f << "zoneA0,";
+		if (sz!=16)
+		{
+			WPS_DEBUG_MSG(("LotusGraph::readGraphZone: the size seems bad for zone 0\n"));
+			f << "###";
+			break;
+		}
+		for (int i=0; i<8; ++i)   // always f0=1, other 0
+		{
+			val=int(libwps::readU16(input));
+			if (val) f << "f" << i << "=" << val << ",";
+		}
+		break;
+	case 4: // seems link to the ref zone, unsure
+		f << "ref,";
+		if (sz!=20)
+		{
+			WPS_DEBUG_MSG(("LotusGraph::readGraphZone: the size seems bad for zone 4\n"));
+			f << "###";
+			break;
+		}
+		for (int i=0; i<10; ++i)   // always f0=1, other 0
+		{
+			val=int(libwps::readU16(input));
+			if (val) f << "f" << i << "=" << val << ",";
+		}
+		break;
+
+	case 0x83: // relative to graphic
+	case 0x84: // rare, when it exists before chartName
+	case 0x86: // rare, when it exists, end of a set. TODO useme
+		f << (id==0x83 ? "GraphBeg" : id==0x84 ? "chartBeg" : "endSet") << ",";
+		if (id==0x86)
+		{
+			if (zId<0 || m_state->m_sheetIdZonePcListMap.find(zId)==m_state->m_sheetIdZonePcListMap.end() ||
+			        m_state->m_sheetIdZonePcListMap.find(zId)->second.m_groupBeginStack.empty())
+			{
+				WPS_DEBUG_MSG(("LotusGraph::readGraphZone: oops can not find the begin of the group\n"));
+				f << "###group,";
+			}
+			else
+			{
+				LotusGraphInternal::ZonePcList &current=m_state->m_sheetIdZonePcListMap.find(zId)->second;
+				size_t begPos=current.m_groupBeginStack.top();
+				if (begPos<current.m_zones.size() && current.m_zones[begPos])
+					current.m_zones[begPos]->m_groupLastPosition=current.m_zones.size();
+				current.m_groupBeginStack.pop();
+			}
+		}
+		if (sz!=0)
+		{
+			WPS_DEBUG_MSG(("LotusGraph::readGraphZone: the size seems bad for zone %d\n", id));
+			f << "###";
+			break;
+		}
+		break;
+	case 0x81: // rare, when it exists, in sheetA4 zones (followed by sheetB5, sheetB6)
+		f << "zoneB1,";
+		if (sz!=6)
+		{
+			WPS_DEBUG_MSG(("LotusGraph::readGraphZone: the size seems bad for zone 81\n"));
+			f << "###";
+			break;
+		}
+		for (int i=0; i<3; ++i)   // always f0=1, other 0
+		{
+			val=int(libwps::read16(input));
+			if (val) f << "f" << i << "=" << val << ",";
+		}
+		break;
+	case 0x95: // relative to graphic
+		f << "zoneB15,";
+		if ((sz%4)!=0)
+		{
+			WPS_DEBUG_MSG(("LotusGraph::readGraphZone: the size seems bad for zone 815\n"));
+			f << "###";
+			break;
+		}
+		if (sz==0) break;
+		val=int(libwps::read16(input));
+		if (m_state->m_actualZonePc)
+			m_state->m_actualZonePc->m_arrows=int(val&3);
+		switch (val&3)
+		{
+		case 1:
+			f << "arrow[beg],";
+			break;
+		case 2:
+			f << "arrow[end],";
+			break;
+		case 3:
+			f << "arrow[beg,end],";
+			break;
+		case 0:
+		default:
+			break;
+		}
+		val &=0xFFFC;
+		if (val) f << "f0=" << val << ",";
+		for (int i=1; i<sz/2; ++i)   // always f0=1, other 0
+		{
+			val=int(libwps::read16(input));
+			if (val) f << "f" << i << "=" << val << ",";
+		}
+		break;
+	case 0x85: // group or ???, after sheetB6.
+	case 0x88: // line after sheetB4
+	case 0x89: // polygon
+	case 0x8a: // freehand
+	case 0x8b: // rect
+	case 0x8c: // ellipse
+	case 0x8d: // arc
+	case 0x8e: // picture
+	case 0x90: // textbox
+	case 0x9a:   // chart, after sheetB4
+	{
+		shared_ptr<LotusGraphInternal::ZonePc> zone(new LotusGraphInternal::ZonePc(stream));
+		if (zId<0)
+		{
+			WPS_DEBUG_MSG(("LotusGraph::readGraphZone: oops can not find the sheet zone id\n"));
+			f << "###sheetId,";
+		}
+		else
+		{
+			// CHECKME: sometimes this does not work, ie. zId is not the spreadssheet data id
+			if (m_state->m_sheetIdZonePcListMap.find(zId)==m_state->m_sheetIdZonePcListMap.end())
+				m_state->m_sheetIdZonePcListMap.insert(std::map<int, LotusGraphInternal::ZonePcList>::value_type(zId, LotusGraphInternal::ZonePcList()));
+			LotusGraphInternal::ZonePcList &current=m_state->m_sheetIdZonePcListMap.find(zId)->second;
+			if (id==0x85) current.m_groupBeginStack.push(current.m_zones.size());
+			current.m_zones.push_back(zone);
+		}
+
+		m_state->m_actualZonePc=zone;
+		int expectedSz=0;
+		switch (id)
+		{
+		case 0x85:
+			zone->m_type=zone->Set;
+			expectedSz=80;
+			break;
+		case 0x88:
+			zone->m_type=zone->Line;
+			expectedSz=70;
+			break;
+		case 0x89:
+			zone->m_type=zone->Polygon;
+			expectedSz=70;
+			break;
+		case 0x8a:
+			zone->m_type=zone->FreeHand;
+			expectedSz=70;
+			break;
+		case 0x8b:
+			zone->m_type=zone->Rect;
+			expectedSz=110;
+			break;
+		case 0x8c:
+			zone->m_type=zone->Ellipse;
+			expectedSz=80;
+			break;
+		case 0x8d:
+			zone->m_type=zone->Arc;
+			expectedSz=84;
+			break;
+		case 0x8e:
+			zone->m_type=zone->Picture;
+			expectedSz=116;
+			break;
+		case 0x90:
+			zone->m_type=zone->TextBox;
+			expectedSz=112;
+			break;
+		case 0x9a:
+			zone->m_type=zone->Chart;
+			expectedSz=126;
+			break;
+		default:
+			break;
+		}
+
+		if (sz<expectedSz)
+		{
+			WPS_DEBUG_MSG(("LotusGraph::readGraphZone: the size seems bad for zone %d\n", id));
+			f << *zone << "###";
+			break;
+		}
+		libwps::DebugStream f2;
+		f2 << "id=" << libwps::read32(input) << ","; // small number or -1
+		for (int i=0; i<5; ++i)   // f2=-1|4
+		{
+			val=int(libwps::readU16(input));
+			if (val==0xFFFF) continue;
+			switch (i)
+			{
+			case 0:
+				if ((val&0x10)==0x10) zone->m_isGroup=true;
+				if ((val&0x40)==0) f2 << "locked,";
+				if ((val&0x200)==0) zone->m_isRoundRect=true;
+				if ((val&0x2000)==0) f2 << "hidden,";
+				val &=0xDDAF;
+				if (val!=0x4d01)
+					f2 << "f0=" << std::hex << val << std::dec << ",";
+				break;
+			case 1:
+				if (val!=0x94)
+					f2 << "f1=" << std::hex << val << std::dec << ",";
+				break;
+			case 2:
+				if ((val>>8)==0x40)
+					zone->m_graphicId[0]=(val&0xFF);
+				else
+				{
+					WPS_DEBUG_MSG(("LotusGraph::readGraphZone: find unexpected graphic style id\n"));
+					f2 << "###GS" << std::hex << val << std::dec << ",";
+				}
+				break;
+			case 3:
+				switch (val)
+				{
+				case 0: // TL and BR
+					break;
+				case 1:
+					f2 << "fasten[TL],";
+					break;
+				case 2:
+					f2 << "no[fasten],";
+					break;
+				default:
+					f2 << "f3=" << val << ",";
+				}
+				break;
+			default:
+			{
+				if (val)
+					f2 << "f" << i << "=" << val << ",";
+				break;
+			}
+			}
+		}
+		int sSz=int(libwps::readU16(input));
+		if ((id!=0x8e && expectedSz+sSz!=sz) || (id==0x8e && expectedSz+sSz>sz))
+		{
+			WPS_DEBUG_MSG(("LotusGraph::readGraphZone: the size seems bad for zone %d\n", id));
+			f << *zone << "###" << f2.str();
+			break;
+		}
+		std::string name;
+		for (int i=0; i<sSz; ++i)
+		{
+			char c=char(libwps::readU8(input));
+			if (c) name+=c;
+			else if (i+1!=sSz)
+			{
+				WPS_DEBUG_MSG(("LotusGraph::readGraphZone: find odd char in zone %d\n", id));
+				f2 << "###";
+			}
+		}
+		if (!name.empty()) f2 << name << ",";
+		f2 << "unkn=[";
+		for (int i=0; i<4; ++i) // often _,3ff00000,_,3ff00000
+		{
+			unsigned long lVal=libwps::readU32(input);
+			if (lVal)
+				f2 << std::hex << lVal << std::dec << ",";
+			else
+				f2 << "_,";
+		}
+		f2 << "],";
+		bool hasFlip=false;
+		for (int i=0; i<10; ++i)
+		{
+			val=int(libwps::readU16(input));
+			if (!val) continue;
+			if (i==0)
+				zone->m_rotate=float(val)/10.f;
+			else if (i==5 && val==1)
+			{
+				hasFlip=true;
+				f2 << "has[flip],";
+			}
+			else
+				f2 << "g" << i << "=" << std::hex << val << std::dec << ",";
+		}
+		float translate[2];
+		for (int i=0; i<2; ++i) translate[i]=unit*float(libwps::read32(input));
+		zone->m_translate=Vec2f(translate[0],translate[1]);
+		for (int i=0; i<2; ++i)
+		{
+			val=int(libwps::readU16(input));
+			if (!val) continue;
+			if (i==0 && hasFlip && val<3)   // useme
+			{
+				if (val&1) f2 << "flipX,";
+				if (val&2) f2 << "flipY,";
+				continue;
+			}
+			f2 << "g" << i+10 << "=" << std::hex << val << std::dec << ",";
+		}
+		if (id>=0x88 && id<=0x8a)
+		{
+			val=int(libwps::read16(input)); // 0
+			if (val) f2 << "h0=" << val << ",";
+			zone->m_numPoints=int(libwps::readU16(input));
+			val=int(libwps::read16(input)); // 0
+			if (val) f2 << "h1=" << val << ",";
+		}
+		else
+		{
+			float dim[4];
+			for (int i=0; i<4; ++i) dim[i]=unit*float(libwps::read32(input));
+			zone->m_box=WPSBox2f(Vec2f(dim[0],dim[1]),Vec2f(dim[2],dim[3]));
+			if (id==0x8b || id==0x8e || id==0x9a)
+			{
+				for (int i=0; i<2; ++i) // h0 small number
+				{
+					val=int(libwps::read16(input));
+					if (val)
+						f2 << "h" << i << "=" << val << ",";
+				}
+				val=int(libwps::readU16(input));
+				if ((val>>8)==0x40)
+					zone->m_graphicId[1]=(val&0xFF);
+				else if (val)
+				{
+					WPS_DEBUG_MSG(("LotusGraph::readGraphZone: find unexpected graphic style id\n"));
+					f2 << "###GS" << std::hex << val << std::dec << ",";
+				}
+			}
+			if (id==0x8d)
+			{
+				for (int i=0; i<2; ++i) // 0
+				{
+					val=int(libwps::read16(input));
+					if (val) f2 << "h" << i+3 << "=" << val << ",";
+				}
+			}
+			else if (id==0x8e || id==0x9a)
+			{
+				for (int i=0; i<11; ++i)   // h3=h4=h6=0|1, h8=h10=0|2d, h13=-1
+				{
+					val=int(libwps::read16(input));
+					if (val) f2 << "h" << i+3 << "=" << val << ",";
+				}
+				if (id==0x8e)
+				{
+					int dSz=int(libwps::readU16(input));
+					if (expectedSz+sSz+dSz!=sz)
+					{
+						WPS_DEBUG_MSG(("LotusGraph::readGraphZone: the size seems bad for zone %d\n", id));
+						f << *zone << "###" << f2.str();
+						break;
+					}
+					std::string dir;
+					for (int i=0; i<dSz; ++i)
+					{
+						char c=char(libwps::readU8(input));
+						if (c) dir+=c;
+						else if (i+1!=dSz)
+						{
+							WPS_DEBUG_MSG(("LotusGraph::readGraphZone: find odd char in zone %d\n", id));
+							f2 << "###";
+						}
+					}
+					if (!dir.empty()) f2 << dir << ",";
+					for (int i=0; i<2; ++i)   // l0=19, l1=24: maybe some zoneId
+					{
+						val=int(libwps::read16(input));
+						if (val) f2 << "l" << i << "=" << val << ",";
+					}
+				}
+				else
+				{
+					f2 << "prev[id]=Z" << int(libwps::read32(input)) << ",";
+					for (int i=0; i<5; ++i)
+					{
+						val=int(libwps::read16(input));
+						int const(expected[])= {0,0,1,1,0};
+						if (val!=expected[i]) f2 << "l" << i << "=" << val << ",";
+					}
+					f2 << "act[id]=Z" << int(libwps::read32(input)) << ",";
+				}
+			}
+		}
+		zone->m_extra=f2.str();
+		f << *zone;
+		break;
+	}
+	default:
+		f << "type=" << std::hex << id << std::dec << ",";
+		break;
+	}
+	ascFile.addPos(pos);
+	ascFile.addNote(f.str().c_str());
+	if (input->tell()!=endPos && input->tell()!=pos)
+		ascFile.addDelimiter(input->tell(),'|');
+	input->seek(endPos, librevenge::RVNG_SEEK_SET);
+	return true;
+}
+
+bool LotusGraph::readGraphDataZone(shared_ptr<WPSStream> stream, long endPos)
+{
+	if (!stream) return false;
+	RVNGInputStreamPtr &input = stream->m_input;
+	libwps::DebugFile &ascFile=stream->m_ascii;
+	libwps::DebugStream f;
+	float const unit=version()>=5 ? 1.f/16.f : 1.f/256.f;
+	f << "Entries(GraphZone)[data]:";
+	long pos = input->tell();
+	int sz=int(endPos-pos);
+	if (m_state->m_actualZonePc && m_state->m_actualZonePc->m_type==LotusGraphInternal::ZonePc::Line && sz==16)
+	{
+		f << "line,";
+		float dim[4];
+		for (int i=0; i<4; ++i) dim[i]=unit*float(libwps::read32(input));
+		m_state->m_actualZonePc->m_box= WPSBox2f(Vec2f(dim[0],dim[1]),Vec2f(dim[2],dim[3]));
+		f << "dim=" << m_state->m_actualZonePc->m_box << ",";
+	}
+	else if (m_state->m_actualZonePc && sz==8*m_state->m_actualZonePc->m_numPoints &&
+	         (m_state->m_actualZonePc->m_type==LotusGraphInternal::ZonePc::FreeHand || m_state->m_actualZonePc->m_type==LotusGraphInternal::ZonePc::Polygon))
+	{
+		f << "poly,pts=[";
+		float dim[2];
+		for (int n=0; n<m_state->m_actualZonePc->m_numPoints; ++n)
+		{
+			for (int i=0; i<2; ++i) dim[i]=unit*float(libwps::read32(input));
+			m_state->m_actualZonePc->m_vertices.push_back(Vec2f(dim[0],dim[1]));
+			f << m_state->m_actualZonePc->m_vertices.back() << ",";
+		}
+		f << "],";
+	}
+	else if (m_state->m_actualZonePc && m_state->m_actualZonePc->m_type==LotusGraphInternal::ZonePc::TextBox)
+	{
+		m_state->m_actualZonePc->m_textBoxEntry.setBegin(pos-2);
+		m_state->m_actualZonePc->m_textBoxEntry.setEnd(endPos);
+		f << "texbox,";
+		std::string text;
+		for (int i=0; i<sz; ++i)
+		{
+			char c=char(libwps::readU8(input));
+			if (c)
+				text+=c;
+			else if (i+1!=sz)
+			{
+				WPS_DEBUG_MSG(("LotusGraph::readGraphDataZone: find unexpected 0 char\n"));
+				f << "###";
+			}
+		}
+		f << text;
+	}
+	else if (m_state->m_actualZonePc && m_state->m_actualZonePc->m_type==LotusGraphInternal::ZonePc::Picture)
+	{
+		/* checkme: the picture is stored in a list on consecutif data zone
+		   and seems preceded by 0100000008000000da0a0000da0a0000de380000
+		 */
+		if (endPos!=pos)
+		{
+			const unsigned char *readData;
+			unsigned long sizeRead, expectedSize=static_cast<unsigned long>(endPos-pos);
+			if ((readData=input->read(expectedSize, sizeRead)) == 0 || sizeRead!=expectedSize)
+			{
+				WPS_DEBUG_MSG(("LotusGraph::readGraphDataZone: can not read the data\n"));
+				f << "###";
+			}
+			else
+			{
+				f << "picture,";
+				if (m_state->m_actualZonePc->m_pictureHeaderRead<20)   // skip the header? zone
+				{
+					int headerRead=m_state->m_actualZonePc->m_pictureHeaderRead+int(expectedSize) > 20 ?
+					               20-m_state->m_actualZonePc->m_pictureHeaderRead : int(expectedSize);
+					m_state->m_actualZonePc->m_pictureHeaderRead+=headerRead;
+					if (headerRead<int(expectedSize))
+					{
+						m_state->m_actualZonePc->m_pictureData.append(readData+headerRead, static_cast<unsigned long>(endPos-pos-headerRead));
+						ascFile.skipZone(pos+headerRead, endPos-1);
+					}
+				}
+				else
+				{
+					m_state->m_actualZonePc->m_pictureData.append(readData, sizeRead);
+					ascFile.skipZone(pos, endPos-1);
+				}
+			}
+		}
+	}
+	else
+	{
+		WPS_DEBUG_MSG(("LotusGraph::readGraphDataZone: find unknown data zone\n"));
+		f << "###";
+	}
+	ascFile.addPos(pos-4);
+	ascFile.addNote(f.str().c_str());
+	if (input->tell()!=endPos)
+		ascFile.addDelimiter(input->tell(),'|');
+	input->seek(endPos, librevenge::RVNG_SEEK_SET);
+	return true;
+}
+
 /* vim:set shiftwidth=4 softtabstop=4 noexpandtab: */
