@@ -34,11 +34,13 @@
 #include "libwps_internal.h"
 #include "libwps_tools_win.h"
 
-#include "WPSCell.h"
 #include "WKSContentListener.h"
+#include "WKSSubDocument.h"
 #include "WPSEntry.h"
 #include "WPSFont.h"
+#include "WPSCell.h"
 #include "WPSStream.h"
+#include "WPSTable.h"
 
 #include "Lotus.h"
 #include "LotusStyleManager.h"
@@ -258,7 +260,7 @@ class Spreadsheet
 public:
 	//! a constructor
 	Spreadsheet() : m_name(""), m_numCols(0), m_numRows(0), m_boundsColsMap(),
-		m_widthColsInChar(), m_rowHeightMap(), m_heightDefault(16),
+		m_widthCols(), m_rowHeightMap(), m_heightDefault(16),
 		m_rowPageBreaksList(), m_positionToCellMap(), m_rowToStyleIdMap(),
 		m_rowToExtraStyleMap() {}
 	//! return a cell corresponding to a spreadsheet, create one if needed
@@ -273,92 +275,97 @@ public:
 		return m_positionToCellMap.find(pos)->second;
 	}
 	//! set the columns size
-	void setColumnWidthInChar(int col, int w=-1)
+	void setColumnWidth(int col, WPSColumnFormat const &format)
 	{
-		if (col < 0) return;
-		if (col >= int(m_widthColsInChar.size()))
+		if (col >= int(m_widthCols.size()))
 		{
 			// sanity check
-			if (col>255 && !m_boundsColsMap.empty() && col >= int(m_widthColsInChar.size())+10 &&
+			if (col>255 && !m_boundsColsMap.empty() && col >= int(m_widthCols.size())+10 &&
 			        m_boundsColsMap.find(col)==m_boundsColsMap.end())
 			{
 				WPS_DEBUG_MSG(("LotusSpreadsheetInternal::Spreadsheet::setColumnWidth: the column %d seems bad\n", col));
 				return;
 			}
-			m_widthColsInChar.resize(size_t(col)+1, -1);
+			WPSColumnFormat defCol;
+			defCol.m_useOptimalWidth=true;
+			m_widthCols.resize(size_t(col)+1, defCol);
 		}
-		m_widthColsInChar[size_t(col)] = w;
+		m_widthCols[size_t(col)] = format;
 		if (col >= m_numCols) m_numCols=col+1;
 	}
 	//! returns the row size in point
-	float getRowHeight(int row) const
+	WPSRowFormat getRowHeight(int row) const
 	{
-		std::map<Vec2i,int>::const_iterator rIt=m_rowHeightMap.lower_bound(Vec2i(-1,row));
+		std::map<Vec2i,WPSRowFormat>::const_iterator rIt=m_rowHeightMap.lower_bound(Vec2i(-1,row));
 		if (rIt!=m_rowHeightMap.end() && rIt->first[0]<=row && rIt->first[1]>=row)
-			return (float) rIt->second;
-		return (float) m_heightDefault;
+			return rIt->second;
+		WPSRowFormat format(m_heightDefault);
+		format.m_isMinimalHeight=true;
+		return format;
 	}
 	//! set the rows size
-	void setRowHeight(int row, int h)
+	void setRowHeight(int row, WPSRowFormat const &format)
 	{
-		if (h>=0)
-			m_rowHeightMap[Vec2i(row,row)]=h;
+		m_rowHeightMap[Vec2i(row,row)]=format;
 	}
 	//! returns the position corresponding to a cell
 	Vec2f getPosition(Vec2i const &cell) const
 	{
 		// first compute the height
 		int lastRow=0;
-		int h=0;
-		std::map<Vec2i,int>::const_iterator rIt=m_rowHeightMap.begin();
+		float h=0;
+		std::map<Vec2i,WPSRowFormat>::const_iterator rIt=m_rowHeightMap.begin();
 		while (rIt!=m_rowHeightMap.end() && rIt->first[1]<cell[1])
 		{
 			if (rIt->first[0]>lastRow)
 			{
-				h+=(rIt->first[0]-lastRow)*m_heightDefault;
+				h+=float(rIt->first[0]-lastRow)*m_heightDefault;
 				lastRow=rIt->first[0];
 			}
-			h+=(rIt->first[1]+1-lastRow)*rIt->second;
+			float rHeight=rIt->second.m_height>=0 ? rIt->second.m_height : m_heightDefault;
+			h+=float(rIt->first[1]+1-lastRow)*rHeight;
 			lastRow=rIt->first[1]+1;
 			++rIt;
 		}
 		if (lastRow<cell[1])
 		{
-			if (rIt!=m_rowHeightMap.end() && rIt->first[0]<cell[1])
-				h+=(cell[1]-lastRow)*rIt->second;
+			if (rIt!=m_rowHeightMap.end() && rIt->first[0]<cell[1] && rIt->second.m_height>=0)
+				h+=float(cell[1]-lastRow)*rIt->second.m_height;
 			else
-				h+=(cell[1]-lastRow)*m_heightDefault;
+				h+=float(cell[1]-lastRow)*m_heightDefault;
 
 		}
 		// now compute the width
-		size_t const numCols = m_widthColsInChar.size();
-		int w=0;
+		size_t const numCols = m_widthCols.size();
+		float w=0;
 		for (size_t i = 0; i < numCols && i<size_t(cell[0]); i++)
-			w+=m_widthColsInChar[i] ? 8*m_widthColsInChar[i] : 72;
+			w+=m_widthCols[i].m_width>=0 ? m_widthCols[i].m_width : 72;
 		if (numCols<size_t(cell[0]))
-			w+=72*int(size_t(cell[0])-numCols);
-		return Vec2f(float(w), float(h));
+			w+=72*float(size_t(cell[0])-numCols);
+		return Vec2f(w, h);
 	}
 	//! try to compress the list of row height
 	void compressRowHeights()
 	{
-		std::map<Vec2i,int> oldMap=m_rowHeightMap;
+		std::map<Vec2i,WPSRowFormat> oldMap=m_rowHeightMap;
 		m_rowHeightMap.clear();
-		std::map<Vec2i,int>::const_iterator rIt=oldMap.begin();
-		int actHeight=-1;
+		std::map<Vec2i,WPSRowFormat>::const_iterator rIt=oldMap.begin();
+		WPSRowFormat actHeight;
+		WPSRowFormat defHeight(m_heightDefault);
+		defHeight.m_isMinimalHeight=true;
 		Vec2i actPos(0,-1);
 		while (rIt!=oldMap.end())
 		{
 			// first check for not filled row
 			if (rIt->first[0]!=actPos[1]+1)
 			{
-				if (actHeight==m_heightDefault)
+				if (actHeight==defHeight)
 					actPos[1]=rIt->first[0]-1;
 				else
 				{
 					if (actPos[1]>=actPos[0])
 						m_rowHeightMap[actPos]=actHeight;
-					actHeight=m_heightDefault;
+					actHeight=defHeight;
 					actPos=Vec2i(actPos[1]+1, rIt->first[0]-1);
 				}
 			}
@@ -375,37 +382,32 @@ public:
 		if (actPos[1]>=actPos[0])
 			m_rowHeightMap[actPos]=actHeight;
 	}
-	//! convert the m_widthColsInChar, m_rowHeightMap in a vector of of point size
-	std::vector<float> getColumnWidths(std::vector<int> &repeated) const
+	//! convert the m_widthColsInChar in a vector of of point size
+	std::vector<WPSColumnFormat> getWidths() const
 	{
-		size_t numElt = m_widthColsInChar.size();
-		std::vector<float> res;
-		size_t actCol=0;
-		int prevWidth=-1;
-		for (size_t i = 0; i < numElt; i++)
+		std::vector<WPSColumnFormat> widths;
+		WPSColumnFormat defWidth, actWidth;
+		defWidth.m_useOptimalWidth=true;
+		int repeat=0;
+		for (size_t c = 0; c < m_widthCols.size(); c++)
 		{
-			int newVal=m_widthColsInChar[i]>=0 ? m_widthColsInChar[i]*8 : 72;
-			if (newVal==prevWidth)
-				continue;
-			if (actCol!=i)
+			WPSColumnFormat newWidth=m_widthCols[c];
+			if (repeat && newWidth!=actWidth)
 			{
-				res.push_back(float(prevWidth));
-				repeated.push_back(int(i-actCol));
+				actWidth.m_numRepeat=repeat;
+				widths.push_back(actWidth);
+				repeat=0;
 			}
-			actCol=i;
-			prevWidth=newVal;
+			if (repeat==0)
+				actWidth=newWidth;
+			++repeat;
 		}
-		if (actCol<numElt)
+		if (repeat)
 		{
-			res.push_back(float(prevWidth));
-			repeated.push_back(int(numElt-actCol));
+			actWidth.m_numRepeat=repeat;
+			widths.push_back(actWidth);
 		}
-		if (numElt<256)
-		{
-			res.push_back(72);
-			repeated.push_back(int(256-numElt));
-		}
-		return res;
+		return widths;
 	}
 	//! returns the row style id corresponding to a sheetId (or -1)
 	int getRowStyleId(int row) const
@@ -429,12 +431,12 @@ public:
 	int m_numRows;
 	/** a map used to stored the min/max row of each columns */
 	std::map<int, Vec2i> m_boundsColsMap;
-	/** the column size in char */
-	std::vector<int> m_widthColsInChar;
+	/** the column size */
+	std::vector<WPSColumnFormat> m_widthCols;
 	/** the map Vec2i(min row, max row) to size in points */
-	std::map<Vec2i,int> m_rowHeightMap;
+	std::map<Vec2i,WPSRowFormat> m_rowHeightMap;
 	/** the default row size in point */
-	int m_heightDefault;
+	float m_heightDefault;
 	/** the list of row page break */
 	std::vector<int> m_rowPageBreaksList;
 	/** a map cell to not empty cells */
@@ -445,12 +447,159 @@ public:
 	std::map<int,ExtraRowStyles> m_rowToExtraStyleMap;
 };
 
+// ------------------------------------------------------------
+// Lotus 123
+// ------------------------------------------------------------
+//! the format style for lotus 123
+struct Format123Style : public WPSCellFormat
+{
+	//! constructor
+	Format123Style() : m_alignAcrossColumn(false)
+	{
+	}
+	//! update the cell style
+	void update(Style &style) const
+	{
+		style.setDTFormat(getFormat(), getDTFormat());
+		style.setFormat(getFormat(), getSubFormat());
+		style.setDigits(digits());
+	}
+	//! operator==
+	bool operator==(Format123Style const &f) const
+	{
+		return m_alignAcrossColumn==f.m_alignAcrossColumn && compare(f)==0;
+	}
+	//! flag to know if we must align across column
+	bool m_alignAcrossColumn;
+};
+
+//! the extra style for lotus 123
+struct Extra123Style
+{
+	//! constructor
+	Extra123Style()
+	{
+		for (int i=0; i<2; ++i)
+			m_borders[i].m_style=WPSBorder::None;
+	}
+	//! returns true if the style is empty
+	bool empty() const
+	{
+		for (int i=0; i<2; ++i)
+		{
+			if (!m_borders[i].isEmpty()) return false;
+		}
+		return true;
+	}
+	//! operator==
+	bool operator==(Extra123Style const &f) const
+	{
+		for (int i=0; i<2; ++i)
+		{
+			if (m_borders[i]!=f.m_borders[i])
+				return false;
+		}
+		return true;
+	}
+	//! update the cell style
+	void update(Style &style) const
+	{
+		for (int i=0; i<2; ++i)
+		{
+			if (m_borders[i].isEmpty()) continue;
+			style.setBorders(i==0 ? WPSBorder::TopBit : WPSBorder::LeftBit,m_borders[i]);
+		}
+	}
+	//! the top/left border
+	WPSBorder m_borders[2];
+};
+
+//! a class used to store the styles of a table in LotusSpreadsheet in a lotus 123 files
+struct Table123Styles
+{
+	//! constructor
+	Table123Styles() : m_defaultCellId(-1), m_rowsToColsToCellIdMap(), m_rowsToColsToExtraStyleMap(), m_rowsToColsToFormatStyleMap()
+	{
+	}
+	//! add a style to a list of cell
+	void addCellStyle(Vec2i const &cols, Vec2i const &rows, int cellId)
+	{
+		if (m_rowsToColsToCellIdMap.find(rows)==m_rowsToColsToCellIdMap.end())
+			m_rowsToColsToCellIdMap[rows]= std::map<Vec2i,int>();
+		std::map<Vec2i,int> &map=m_rowsToColsToCellIdMap.find(rows)->second;
+		if (map.find(cols)!=map.end())
+		{
+			WPS_DEBUG_MSG(("LotusSpreadsheetInternal::Table123Styles::addCellStyle: find dupplicated cell\n"));
+		}
+		std::map<Vec2i,int>::iterator it=map.lower_bound(cols);
+		if (!map.empty() && it!=map.begin()) --it;
+		if (it!=map.end() && it->first[1]+1==cols[0] && it->second==cellId)
+		{
+			Vec2i newCols=it->first;
+			map.erase(newCols);
+			newCols[1]=cols[1];
+			map[newCols]=cellId;
+		}
+		else
+			map[cols]=cellId;
+	}
+	//! add a extra style to a list of cell
+	void addCellStyle(Vec2i const &cols, Vec2i const &rows, Extra123Style const &extra)
+	{
+		if (m_rowsToColsToExtraStyleMap.find(rows)==m_rowsToColsToExtraStyleMap.end())
+			m_rowsToColsToExtraStyleMap[rows]= std::map<Vec2i,Extra123Style>();
+		std::map<Vec2i,Extra123Style> &map=m_rowsToColsToExtraStyleMap.find(rows)->second;
+		// checkme: sometimes, we can retrieve the same cells again
+		std::map<Vec2i,Extra123Style>::iterator it=map.lower_bound(cols);
+		if (!map.empty() && it!=map.begin()) --it;
+		if (it!=map.end() && it->first[1]+1==cols[0] && it->second==extra)
+		{
+			Vec2i newCols=it->first;
+			map.erase(newCols);
+			newCols[1]=cols[1];
+			map[newCols]=extra;
+		}
+		else
+			map[cols]=extra;
+	}
+	//! add a extra style to a list of cell
+	void addCellStyle(Vec2i const &cols, Vec2i const &rows, Format123Style const &format)
+	{
+		if (m_rowsToColsToFormatStyleMap.find(rows)==m_rowsToColsToFormatStyleMap.end())
+			m_rowsToColsToFormatStyleMap[rows]= std::map<Vec2i,Format123Style>();
+		std::map<Vec2i,Format123Style> &map=m_rowsToColsToFormatStyleMap.find(rows)->second;
+		if (map.find(cols)!=map.end())
+		{
+			WPS_DEBUG_MSG(("LotusSpreadsheetInternal::Table123Styles::addCellStyle: find dupplicated cell\n"));
+		}
+		std::map<Vec2i,Format123Style>::iterator it=map.lower_bound(cols);
+		if (!map.empty() && it!=map.begin()) --it;
+		if (it!=map.end() && it->first[1]+1==cols[0] && it->second==format)
+		{
+			Vec2i newCols=it->first;
+			map.erase(newCols);
+			newCols[1]=cols[1];
+			map[newCols]=format;
+		}
+		else
+			map[cols]=format;
+	}
+	//! the default cell style
+	int m_defaultCellId;
+	//! map rows to cols to cell id
+	std::map<Vec2i, std::map<Vec2i,int> > m_rowsToColsToCellIdMap;
+	//! map rows to cols to extra style
+	std::map<Vec2i, std::map<Vec2i,Extra123Style> > m_rowsToColsToExtraStyleMap;
+	//! map rows to cols to format style
+	std::map<Vec2i, std::map<Vec2i,Format123Style> > m_rowsToColsToFormatStyleMap;
+};
+
 //! the state of LotusSpreadsheet
 struct State
 {
 	//! constructor
 	State() :  m_version(-1), m_spreadsheetList(), m_nameToCellsMap(),
-		m_rowStylesList(), m_rowSheetIdToStyleIdMap(), m_rowSheetIdToChildRowIdMap(),
+		m_rowStylesList(), m_rowSheetIdToStyleIdMap(), m_rowSheetIdToChildRowIdMap(), m_sheetIdToTableStyleMap(),
 		m_sheetCurrentId(-1)
 	{
 		m_spreadsheetList.resize(1);
@@ -465,12 +614,53 @@ struct State
 	{
 		if (id<0||id>=int(m_spreadsheetList.size()))
 		{
-			WPS_DEBUG_MSG(("LotusSpreadsheetInternal::State::getSheet: can find spreadsheet %d\n", id));
+			WPS_DEBUG_MSG(("LotusSpreadsheetInternal::State::getSheet: can not find spreadsheet %d\n", id));
 			static Spreadsheet empty;
 			return empty;
 		}
 		return m_spreadsheetList[size_t(id)];
 	}
+	//! returns a table style for a sheet(if it exists)
+	Table123Styles const *getTableStyle(int id) const
+	{
+		Vec2i pos(-1,id);
+		std::map<Vec2i,Table123Styles>::const_iterator cIt=m_sheetIdToTableStyleMap.lower_bound(pos);
+		if (cIt==m_sheetIdToTableStyleMap.end() || cIt->first[0]>id || cIt->first[1]<id)
+			return 0;
+		return &cIt->second;
+	}
+	//! returns a table style for a sheet zone, create it if needed
+	Table123Styles *getTablesStyle(Vec2i pos)
+	{
+		std::map<Vec2i,Table123Styles>::iterator cIt=m_sheetIdToTableStyleMap.lower_bound(pos);
+		if (cIt==m_sheetIdToTableStyleMap.end() || cIt->first[0]>pos[1] || cIt->first[1]<pos[0])
+		{
+			m_sheetIdToTableStyleMap[pos]=Table123Styles();
+			return &m_sheetIdToTableStyleMap.find(pos)->second;
+		}
+		if (cIt->first==pos)
+			return &cIt->second;
+		Vec2i actPos=cIt->first;
+		if (actPos[0]>pos[0] || actPos[1]<pos[1])
+		{
+			WPS_DEBUG_MSG(("LotusSpreadsheetInternal::State::getTablesStyle: problem when creating spreadsheet %d,%d\n", pos[0], pos[1]));
+			return 0;
+		}
+		Table123Styles const &table=cIt->second;
+		if (actPos[0]<pos[0])
+		{
+			Vec2i newPos(actPos[0],pos[0]);
+			m_sheetIdToTableStyleMap[newPos]=table;
+		}
+		m_sheetIdToTableStyleMap[pos]=table;
+		if (actPos[1]>pos[1])
+		{
+			Vec2i newPos(pos[1],actPos[1]);
+			m_sheetIdToTableStyleMap[newPos]=table;
+		}
+		m_sheetIdToTableStyleMap.erase(actPos);
+		return &m_sheetIdToTableStyleMap.find(pos)->second;
+	}
 	//! returns the ith spreadsheet
 	librevenge::RVNGString getSheetName(int id) const
 	{
@@ -492,10 +682,55 @@ struct State
 	std::map<Vec2i,size_t> m_rowSheetIdToStyleIdMap;
 	//! map Vec2i(row, sheetId) to child style
 	std::multimap<Vec2i,Vec2i> m_rowSheetIdToChildRowIdMap;
+	//! map Vec2i(sheetMin, sheetMax) to table style
+	std::map<Vec2i,Table123Styles> m_sheetIdToTableStyleMap;
 	//! the sheet id
 	int m_sheetCurrentId;
 };
 
+//! Internal: the subdocument of a LotusSpreadsheet
+class SubDocument : public WKSSubDocument
+{
+public:
+	//! constructor for a text entry
+	SubDocument(RVNGInputStreamPtr input, LotusSpreadsheet &sheetParser, WPSEntry const &entry) :
+		WKSSubDocument(input, 0), m_sheetParser(&sheetParser), m_entry(entry) {}
+	//! destructor
+	~SubDocument() {}
+
+	//! operator==
+	virtual bool operator==(shared_ptr<WKSSubDocument> const &doc) const
+	{
+		if (!doc || !WKSSubDocument::operator==(doc))
+			return false;
+		SubDocument const *sDoc = dynamic_cast<SubDocument const *>(doc.get());
+		if (!sDoc) return false;
+		if (m_sheetParser != sDoc->m_sheetParser) return false;
+		if (m_entry != sDoc->m_entry) return false;
+		return true;
+	}
+
+	//! the parser function
+	void parse(shared_ptr<WKSContentListener> &listener, libwps::SubDocumentType subDocumentType);
+	//! the spreadsheet parse
+	LotusSpreadsheet *m_sheetParser;
+	//! a text zone entry
+	WPSEntry m_entry;
+private:
+	SubDocument(SubDocument const &);
+	SubDocument &operator=(SubDocument const &);
+};
+
+void SubDocument::parse(shared_ptr<WKSContentListener> &listener, libwps::SubDocumentType)
+{
+	if (!m_sheetParser)
+	{
+		listener->insertCharacter(' ');
+		WPS_DEBUG_MSG(("LotusSpreadsheetInternal::SubDocument::parse: bad parser\n"));
+		return;
+	}
+	m_sheetParser->sendTextNote(m_input, m_entry);
+}
 }
 
 // constructor, destructor
@@ -757,7 +992,7 @@ bool LotusSpreadsheet::readColumnSizes(shared_ptr<WPSStream> stream)
 	{
 		int col=(int)libwps::readU8(input);
 		int width=(int)libwps::readU8(input); // width in char, default 12...
-		sheet->setColumnWidthInChar(col, width);
+		sheet->setColumnWidth(col, WPSColumnFormat(float(7*width)));
 		f << width << "C:col" << col << ",";
 	}
 	f << "],";
@@ -1064,7 +1299,7 @@ bool LotusSpreadsheet::readRowFormat(shared_ptr<WPSStream> stream, LotusSpreadsh
 	}
 	else if (wh&0x10)
 	{
-		int fId=(value[1]>>6)&0xFF;
+		int fId=(value[1]>>6);
 		if (fId==0)
 			;
 		else if ((wh&0xf)==0)
@@ -1076,7 +1311,7 @@ bool LotusSpreadsheet::readRowFormat(shared_ptr<WPSStream> stream, LotusSpreadsh
 		}
 		else
 			f << "#fId=" << fId << ",";
-		value[1] &= 0x303C;
+		value[1] &= 0x3C;
 	}
 	else
 	{
@@ -1100,6 +1335,359 @@ bool LotusSpreadsheet::readRowFormat(shared_ptr<WPSStream> stream, LotusSpreadsh
 	return true;
 }
 
+bool LotusSpreadsheet::readCellsFormat801(shared_ptr<WPSStream> stream, WPSVec3i const &minC, WPSVec3i const &maxC, int subZoneId)
+{
+	if (!stream) return false;
+	RVNGInputStreamPtr &input=stream->m_input;
+	libwps::DebugFile &ascFile=stream->m_ascii;
+	libwps::DebugStream f;
+
+	long pos = input->tell();
+	long type = (long) libwps::read16(input);
+	if (type != 0x801)
+	{
+		WPS_DEBUG_MSG(("LotusSpreadsheet::readCellsFormat801: not a cells formats definition\n"));
+		return false;
+	}
+	long sz = (long) libwps::readU16(input);
+	long endPos = pos+4+sz;
+	if (sz<0 || !stream->checkFilePosition(endPos))
+	{
+		input->seek(pos, librevenge::RVNG_SEEK_SET);
+		return false;
+	}
+	f << "Entries(Zone8):";
+	if (minC==maxC)
+		f << "setStyle[" << minC << "],";
+	else
+		f << "setStyle[" << minC << "<->" << maxC << "],";
+	LotusSpreadsheetInternal::Table123Styles *tableStyle=0;
+	int const vers=version();
+	if (minC[0]<=maxC[0] && minC[0]>=0)
+		tableStyle=m_state->getTablesStyle(Vec2i(minC[0],maxC[0]));
+
+	int val;
+	switch (sz)
+	{
+	case 2: // in general the second one
+		val=int(libwps::readU16(input));
+		if ((val>>8)==0x50)
+		{
+			if (tableStyle) tableStyle->addCellStyle(Vec2i(minC[1],maxC[1]),Vec2i(minC[2],maxC[2]),(val&0xFF));
+			f << "Ce" << (val&0xFF);
+		}
+		else
+		{
+			WPS_DEBUG_MSG(("LotusSpreadsheet::readCellsFormat801: find unexpected format type\n"));
+			f << "##" << std::hex << val << std::dec << ",";
+		}
+		break;
+	case 4:   // in general the first one
+	{
+		LotusSpreadsheetInternal::Format123Style format;
+		int values[4];
+		for (int i=0; i<4; ++i) values[i]=int(libwps::readU8(input));
+		if ((values[0]&0x80)==0)
+			f << "protected[no],";
+		else
+			values[0] &= 0x7f;
+		if ((values[3]&0x80) && (values[0]>>4)==0)
+		{
+			if (values[2]>=0x7a)
+				format.setDTFormat(WPSCellFormat::F_TIME, "%I:%M:%S%p");
+			else if (values[2]>=0x63)
+				format.setDTFormat(WPSCellFormat::F_DATE, "%m/%d/%y");
+			else
+			{
+				format.setFormat(WPSCellFormat::F_NUMBER, 4);
+				format.setDigits(values[0]&0xF);
+			}
+			values[3]&=0x7f;
+		}
+		else if (values[0]!=0x7f)
+		{
+			switch ((values[0]>>4))
+			{
+			case 0: // fixed
+				f << "fixed,";
+				format.setFormat(WPSCellFormat::F_NUMBER, 1);
+				format.setDigits(values[0]&0xF);
+				break;
+			case 1: // scientific
+				format.setFormat(WPSCellFormat::F_NUMBER, 2);
+				format.setDigits(values[0]&0xF);
+				break;
+			case 2: // currency
+				format.setFormat(WPSCellFormat::F_NUMBER, 4);
+				format.setDigits(values[0]&0xF);
+				break;
+			case 3: // percent
+				format.setFormat(WPSCellFormat::F_NUMBER, 3);
+				format.setDigits(values[0]&0xF);
+				break;
+			case 4: // decimal
+				format.setFormat(WPSCellFormat::F_NUMBER, 1);
+				format.setDigits(values[0]&0xF);
+				break;
+			case 7: // checkme
+				switch (values[0]&0xF)
+				{
+				case 0: // +/- : kind of bool
+					format.setFormat(WPSCellFormat::F_BOOLEAN);
+					f << "+/-,";
+					break;
+				case 1:
+					format.setFormat(WPSCellFormat::F_NUMBER, 0);
+					break;
+				case 2:
+					format.setDTFormat(WPSCellFormat::F_DATE, "%d %B %y");
+					break;
+				case 3:
+					format.setDTFormat(WPSCellFormat::F_DATE, "%d %B");
+					break;
+				case 4:
+					format.setDTFormat(WPSCellFormat::F_DATE, "%B %y");
+					break;
+				case 5:
+					format.setFormat(WPSCellFormat::F_TEXT);
+					break;
+				case 6:
+					format.setFormat(WPSCellFormat::F_TEXT);
+					// font.m_attributes |= WPS_HIDDEN_BIT;
+					break;
+				case 7:
+					format.setDTFormat(WPSCellFormat::F_TIME, "%I:%M:%S%p");
+					break;
+				case 8:
+					format.setDTFormat(WPSCellFormat::F_TIME, "%I:%M%p");
+					break;
+				case 9:
+					format.setDTFormat(WPSCellFormat::F_DATE, "%m/%d/%y");
+					break;
+				case 0xa:
+					format.setDTFormat(WPSCellFormat::F_DATE, "%m/%d");
+					break;
+				case 0xb:
+					format.setDTFormat(WPSCellFormat::F_TIME, "%H:%M:%S");
+					break;
+				case 0xc:
+					format.setDTFormat(WPSCellFormat::F_TIME, "%H:%M");
+					break;
+				case 0xd:
+					format.setFormat(WPSCellFormat::F_TEXT);
+					f << "label,";
+					break;
+				default:
+					WPS_DEBUG_MSG(("LotusSpreadsheet::readCellsFormat801: find unknown 7e format\n"));
+					f << "Fo=##7e,";
+					break;
+				}
+				break;
+			default:
+				WPS_DEBUG_MSG(("LotusSpreadsheet::readCellsFormat801: find unknown %x format\n", (unsigned int)(values[0])));
+				f << "##Fo=" << std::hex << values[0] << std::dec << ",";
+				break;
+			}
+		}
+		f << format << ",";
+
+		if (values[1]&1) f << "neg[value,red],";
+		if (values[1]&2) f << "add[parenthesis],";
+		if (values[1]&0x10)
+		{
+			format.m_alignAcrossColumn=true;
+			f << "align[across,column],";
+		}
+		if (values[1]&0x20) f << "hidden,";
+		values[1] &=0xCC;
+		for (int i=1; i<4; ++i)
+		{
+			if (values[i]) f << "f" << i << "=" << std::hex << values[i] << std::dec << ",";
+		}
+		if (tableStyle) tableStyle->addCellStyle(Vec2i(minC[1],maxC[1]),Vec2i(minC[2],maxC[2]),format);
+		break;
+	}
+	case 8:   // in general the third one, this define border, ...
+	{
+		LotusSpreadsheetInternal::Extra123Style style;
+		for (int i=0; i<2; ++i)
+		{
+			int col=int(libwps::readU8(input));
+			val=int(libwps::readU8(input));
+			if ((val&0xF)==0xF) continue;
+			f << (i==0 ?  "bordT" : "bordL") << "=[";
+			WPSBorder border;
+			switch (val&0xF)
+			{
+			case 0:
+				border.m_style=WPSBorder::None;
+				break;
+			case 1: // single
+				break;
+			case 2:
+				border.m_type=WPSBorder::Double;
+				break;
+			case 3:
+				border.m_width=2;
+				break;
+			case 4: // dot[1x1]
+				border.m_style=WPSBorder::Dot;
+				break;
+			case 5:
+				border.m_style=WPSBorder::Dash;
+				f << "dash[1x3],";
+				break;
+			case 6:
+				border.m_style=WPSBorder::Dash;
+				f << "dash[3x1],";
+				break;
+			case 7:
+				border.m_style=WPSBorder::Dash;
+				f << "dash[1x1,3x1],";
+				break;
+			case 8:
+				border.m_style=WPSBorder::Dash;
+				f << "dash[1x1,1x1,2x1],";
+				break;
+			default:
+				f << "##type=" << (val&0xF) << ",";
+				break;
+			}
+			if (!m_styleManager->getColor256(col,border.m_color))
+				f << "##colId=" << col << ",";
+			f << border;
+			f << "],";
+			style.m_borders[i]=border;
+		}
+		f << "unk0=[";
+		for (int i=0; i<4; ++i)
+		{
+			val=int(libwps::readU8(input));
+			if (val)
+				f <<std::hex << val << std::dec << ",";
+			else
+				f << "_,";
+		}
+		f << "],";
+		if (tableStyle) tableStyle->addCellStyle(Vec2i(minC[1],maxC[1]),Vec2i(minC[2],maxC[2]),style);
+		break;
+	}
+	case 12:   // column data, in general the fourst one
+	{
+		if (subZoneId==0) f << "col,";
+		else if (subZoneId==1) f << "row,";
+		else if (subZoneId!=-1)
+		{
+			WPS_DEBUG_MSG(("LotusSpreadsheet::readCellsFormat801: the zone8 id seems bad\n"));
+			f << "###zone15=" << subZoneId << ",";
+		}
+		int width=-1;
+		bool isDefault=false;
+		bool isWidthDef=true;
+		f << "colUnkn=[";
+		for (int i=0; i<7; ++i)   // f1=[01][04][01], f2: big number, f3=0|1, f4=0|2d|3c|240
+		{
+			if (i==2 && vers>=5) continue;
+			val=(vers<5 && (i==2||i==4)) ? int(libwps::readU8(input)) : int(libwps::readU16(input));
+			if (i==1)
+			{
+				if ((val&1)==0)
+				{
+					isDefault=true;
+					f << "no[w],";
+				}
+				if (val&2) f << "hidden,";
+				if (val&0x20) f << "page[break],";
+				if (val&0x40)
+				{
+					isWidthDef=false;
+					f << "w[def],";
+				}
+				if (val&0x100)
+					f << "fl100,";
+				val &= 0xFE9C;
+				if (val)
+					f << "##fl=" << std::hex << val << std::dec << ",";
+			}
+			else if (i==3)
+			{
+				width=val;
+				if (!isDefault)
+					f << "w=" << width << ",";
+			}
+			else if (val)
+				f <<std::hex << val << std::dec << ",";
+			else
+				f << "_,";
+		}
+		f << "],";
+		if (isDefault || minC[1]<0 || width<0) break;
+		if (subZoneId<0 || subZoneId>1) break;
+		for (int i=minC[0]; i<=maxC[0]; ++i)
+		{
+			LotusSpreadsheetInternal::Spreadsheet &sheet=m_state->getSheet(i);
+			for (int c=minC[1]; c<=maxC[1]; ++c)
+			{
+				if (subZoneId==0)
+				{
+					WPSColumnFormat format(!isWidthDef ? 72 : vers>=5 ? float(width)/16.f : float(width));
+					sheet.setColumnWidth(c, format);
+				}
+				else
+				{
+					WPSRowFormat format(vers>=5 ? float(width)/16.f : float(width));
+					format.m_useOptimalHeight=!isWidthDef;
+					sheet.setRowHeight(c, format);
+				}
+			}
+		}
+		break;
+	}
+	case 30: // table data, rare, the last one
+		for (int i=0; i<2; ++i)   // f0=71|7e
+		{
+			val=int(libwps::readU16(input));
+			if (val)
+				f << "f" << i << "=" << val << ",";
+		}
+		f << "],";
+		val=int(libwps::readU16(input));
+		if ((val>>8)==0x50)
+		{
+			if (tableStyle) tableStyle->m_defaultCellId=(val&0xFF);
+			f << "Ce" << (val&0xFF) << ",";
+		}
+		else
+		{
+			WPS_DEBUG_MSG(("LotusSpreadsheet::readCellsFormat801: find unexpected format type\n"));
+			f << "##" << std::hex << val << std::dec << ",";
+		}
+		f << "tableUnk=[";
+		for (int i=0; i<12; ++i)
+		{
+			// f0=168|438|3000|3600, f2=f0|14a|a008|c00,f4=[02]0[04c]0,f6=0|3
+			// f8=[37]5d4,f10=0|4,f11=1003(L3?)
+			val=int(libwps::readU16(input));
+			if (val)
+				f <<std::hex << val << std::dec << ",";
+			else
+				f << "_,";
+		}
+		f << "],";
+		break;
+	default:
+		break;
+	}
+	ascFile.addPos(pos);
+	ascFile.addNote(f.str().c_str());
+	if (input->tell()!=endPos)
+	{
+		ascFile.addDelimiter(input->tell(),'|');
+		input->seek(endPos, librevenge::RVNG_SEEK_SET);
+	}
+	return true;
+}
+
 bool LotusSpreadsheet::readRowSizes(shared_ptr<WPSStream> stream, long endPos)
 {
 	if (!stream) return false;
@@ -1146,7 +1734,7 @@ bool LotusSpreadsheet::readRowSizes(shared_ptr<WPSStream> stream, long endPos)
 		if (val!=0xFFFF)
 		{
 			f << "dim=" << float(val+31)/32.f << ",";
-			sheet->setRowHeight(row, int((val+31)/32));
+			sheet->setRowHeight(row, WPSRowFormat(float(val+31)/32.f));
 		}
 		for (int j=0; j<2; ++j)
 		{
@@ -1294,7 +1882,7 @@ bool LotusSpreadsheet::readExtraRowFormats(shared_ptr<WPSStream> stream)
 
 	LotusSpreadsheetInternal::Spreadsheet &sheet=m_state->getSheet(m_state->m_sheetCurrentId);
 	int val=int(libwps::readU8(input));
-	sheet.setRowHeight(row, val);
+	sheet.setRowHeight(row, WPSRowFormat(float(val)));
 	if (val!=14) f << "height=" << val << ",";
 	val=int(libwps::readU8(input)); // 10|80
 	if (val) f << "f0=" << std::hex << val << std::dec << ",";
@@ -1769,10 +2357,7 @@ void LotusSpreadsheet::sendSpreadsheet(int sheetId)
 		return;
 	}
 	LotusSpreadsheetInternal::Spreadsheet &sheet = m_state->getSheet(sheetId);
-	std::vector<int> repeated;
-	std::vector<float> colWidths=sheet.getColumnWidths(repeated);
-	m_listener->openSheet(colWidths, librevenge::RVNG_POINT,
-	                      repeated, m_state->getSheetName(sheetId));
+	m_listener->openSheet(sheet.getWidths(), m_state->getSheetName(sheetId));
 	m_mainParser.sendGraphics(sheetId);
 	sheet.compressRowHeights();
 	/* create a set to know which row needed to be send, each value of
@@ -1804,7 +2389,7 @@ void LotusSpreadsheet::sendSpreadsheet(int sheetId)
 		newRowSet.insert(rows[0]);
 		newRowSet.insert(rows[1]+1);
 	}
-	for (std::map<Vec2i,int>::const_iterator rIt=sheet.m_rowHeightMap.begin();
+	for (std::map<Vec2i,WPSRowFormat>::const_iterator rIt=sheet.m_rowHeightMap.begin();
 	        rIt!=sheet.m_rowHeightMap.end(); ++rIt)
 	{
 		Vec2i rows=rIt->first;
@@ -1819,6 +2404,31 @@ void LotusSpreadsheet::sendSpreadsheet(int sheetId)
 		newRowSet.insert(rIt->first+1);
 	}
 
+	LotusSpreadsheetInternal::Table123Styles const *table123Styles=m_state->getTableStyle(sheetId);
+	if (table123Styles)
+	{
+		for (std::map<Vec2i, std::map<Vec2i,int> >::const_iterator rIt=table123Styles->m_rowsToColsToCellIdMap.begin();
+		        rIt!=table123Styles->m_rowsToColsToCellIdMap.end(); ++rIt)
+		{
+			Vec2i rows=rIt->first;
+			newRowSet.insert(rows[0]);
+			newRowSet.insert(rows[1]+1);
+		}
+		for (std::map<Vec2i, std::map<Vec2i,LotusSpreadsheetInternal::Extra123Style> >::const_iterator rIt=table123Styles->m_rowsToColsToExtraStyleMap.begin();
+		        rIt!=table123Styles->m_rowsToColsToExtraStyleMap.end(); ++rIt)
+		{
+			Vec2i rows=rIt->first;
+			newRowSet.insert(rows[0]);
+			newRowSet.insert(rows[1]+1);
+		}
+		for (std::map<Vec2i, std::map<Vec2i,LotusSpreadsheetInternal::Format123Style> >::const_iterator rIt=table123Styles->m_rowsToColsToFormatStyleMap.begin();
+		        rIt!=table123Styles->m_rowsToColsToFormatStyleMap.end(); ++rIt)
+		{
+			Vec2i rows=rIt->first;
+			newRowSet.insert(rows[0]);
+			newRowSet.insert(rows[1]+1);
+		}
+	}
 	for (std::set<int>::const_iterator sIt=newRowSet.begin(); sIt!=newRowSet.end();)
 	{
 		int row=*(sIt++);
@@ -1829,14 +2439,14 @@ void LotusSpreadsheet::sendSpreadsheet(int sheetId)
 		}
 		if (sIt==newRowSet.end())
 			break;
-		m_listener->openSheetRow(sheet.getRowHeight(row), librevenge::RVNG_POINT, false, *sIt-row);
-		sendRowContent(sheet, row);
+		m_listener->openSheetRow(sheet.getRowHeight(row), *sIt-row);
+		sendRowContent(sheet, row, table123Styles);
 		m_listener->closeSheetRow();
 	}
 	m_listener->closeSheet();
 }
 
-void LotusSpreadsheet::sendRowContent(LotusSpreadsheetInternal::Spreadsheet const &sheet, int row)
+void LotusSpreadsheet::sendRowContent(LotusSpreadsheetInternal::Spreadsheet const &sheet, int row, LotusSpreadsheetInternal::Table123Styles const *table123Styles)
 {
 	if (m_listener.get() == 0L)
 	{
@@ -1899,8 +2509,72 @@ void LotusSpreadsheet::sendRowContent(LotusSpreadsheetInternal::Spreadsheet cons
 			extraStyles=0;
 	}
 
+	libwps_tools_win::Font::Type const defFontType=m_mainParser.getDefaultFontType();
+	LotusSpreadsheetInternal::Style defaultStyle(defFontType);
+
+	std::map<Vec2i, LotusSpreadsheetInternal::Style> colToCellIdMap;
+	std::map<Vec2i, LotusSpreadsheetInternal::Style>::const_iterator c123It;
+	std::map<Vec2i, LotusSpreadsheetInternal::Extra123Style> colToExtraStyleMap;
+	std::map<Vec2i, LotusSpreadsheetInternal::Extra123Style>::const_iterator e123It;
+	std::map<Vec2i, LotusSpreadsheetInternal::Format123Style> colToFormatStyleMap;
+	std::map<Vec2i, LotusSpreadsheetInternal::Format123Style>::const_iterator f123It;
+
+	std::map<int,int> potentialMergeMap;
+	if (table123Styles)
+	{
+		for (std::map<Vec2i, std::map<Vec2i,int> >::const_iterator rIt=table123Styles->m_rowsToColsToCellIdMap.begin();
+		        rIt!=table123Styles->m_rowsToColsToCellIdMap.end(); ++rIt)
+		{
+			if (rIt->first[0]>row || rIt->first[1]<row) continue;
+			for (std::map<Vec2i,int>::const_iterator colIt=rIt->second.begin(); colIt!=rIt->second.end(); ++colIt)
+			{
+				LotusSpreadsheetInternal::Style style(defFontType);
+				WPSFont font;
+				if (!m_styleManager->updateCellStyle(colIt->second, style, font, style.m_fontType))
+					continue;
+				style.setFont(font);
+				colToCellIdMap.insert(std::map<Vec2i, LotusSpreadsheetInternal::Style>::value_type(colIt->first,style));
+				newColSet.insert(colIt->first[0]);
+				newColSet.insert(colIt->first[1]+1);
+			}
+		}
+		for (std::map<Vec2i, std::map<Vec2i, LotusSpreadsheetInternal::Extra123Style> >::const_iterator rIt=table123Styles->m_rowsToColsToExtraStyleMap.begin();
+		        rIt!=table123Styles->m_rowsToColsToExtraStyleMap.end(); ++rIt)
+		{
+			if (rIt->first[0]>row || rIt->first[1]<row) continue;
+			for (std::map<Vec2i,LotusSpreadsheetInternal::Extra123Style>::const_iterator colIt=rIt->second.begin(); colIt!=rIt->second.end(); ++colIt)
+			{
+				colToExtraStyleMap.insert(std::map<Vec2i, LotusSpreadsheetInternal::Extra123Style>::value_type(colIt->first,colIt->second));
+				newColSet.insert(colIt->first[0]);
+				newColSet.insert(colIt->first[1]+1);
+			}
+		}
+		for (std::map<Vec2i, std::map<Vec2i, LotusSpreadsheetInternal::Format123Style> >::const_iterator rIt=table123Styles->m_rowsToColsToFormatStyleMap.begin();
+		        rIt!=table123Styles->m_rowsToColsToFormatStyleMap.end(); ++rIt)
+		{
+			if (rIt->first[0]>row || rIt->first[1]<row) continue;
+			for (std::map<Vec2i,LotusSpreadsheetInternal::Format123Style>::const_iterator colIt=rIt->second.begin(); colIt!=rIt->second.end(); ++colIt)
+			{
+				colToFormatStyleMap.insert(std::map<Vec2i, LotusSpreadsheetInternal::Format123Style>::value_type(colIt->first,colIt->second));
+				newColSet.insert(colIt->first[0]);
+				newColSet.insert(colIt->first[1]+1);
+				if (colIt->first[0]!=colIt->first[1] && colIt->second.m_alignAcrossColumn)
+					potentialMergeMap[colIt->first[0]]=colIt->first[1]+1;
+			}
+		}
+		c123It=colToCellIdMap.begin();
+		e123It=colToExtraStyleMap.begin();
+		f123It=colToFormatStyleMap.begin();
+		if (table123Styles->m_defaultCellId>=0)
+		{
+			WPSFont font;
+			if (m_styleManager->updateCellStyle(table123Styles->m_defaultCellId, defaultStyle, font, defaultStyle.m_fontType))
+				defaultStyle.setFont(font);
+		}
+		if (colToCellIdMap.empty() && colToExtraStyleMap.empty() && colToFormatStyleMap.empty())
+			table123Styles=0;
+	}
 
-	LotusSpreadsheetInternal::Style defaultStyle(m_mainParser.getDefaultFontType());
 	LotusSpreadsheetInternal::Cell emptyCell;
 	for (std::set<int>::const_iterator colIt=newColSet.begin(); colIt!=newColSet.end();)
 	{
@@ -1935,6 +2609,34 @@ void LotusSpreadsheet::sendRowContent(LotusSpreadsheetInternal::Spreadsheet cons
 		}
 		LotusSpreadsheetInternal::Style style=styles ? sIt->second : defaultStyle;
 		bool hasStyle=styles!=0;
+		if (table123Styles)
+		{
+			while (c123It!=colToCellIdMap.end() && c123It->first[1] < col)
+				++c123It;
+			if (c123It!=colToCellIdMap.end() && c123It->first[0] <= col && c123It->first[1] >= col)
+			{
+				style=c123It->second;
+				hasStyle=true;
+			}
+
+			while (e123It!=colToExtraStyleMap.end() && e123It->first[1] < col)
+				++e123It;
+			if (e123It!=colToExtraStyleMap.end() && e123It->first[0] <= col && e123It->first[1] >= col)
+			{
+				e123It->second.update(style);
+				hasStyle=true;
+			}
+			while (f123It!=colToFormatStyleMap.end() && f123It->first[1] < col)
+				++f123It;
+			if (f123It!=colToFormatStyleMap.end() && f123It->first[0] <= col && f123It->first[1] >= col)
+			{
+				f123It->second.update(style);
+				hasStyle=true;
+			}
+			if (c123It==colToCellIdMap.end() && e123It==colToExtraStyleMap.end() && f123It==colToFormatStyleMap.end())
+				table123Styles=0;
+		}
+
 		if (extraStyles && !eIt->second.empty())
 		{
 			eIt->second.update(style);
@@ -1957,7 +2659,28 @@ void LotusSpreadsheet::sendRowContent(LotusSpreadsheetInternal::Spreadsheet cons
 		}
 		if (!hasCell && !hasStyle) continue;
 		if (!hasCell) emptyCell.setPosition(Vec2i(col, row));
-		sendCellContent(hasCell ? cIt->second : emptyCell, style, endCol-col);
+		bool canMerge=false;
+		if (hasCell && potentialMergeMap.find(col)!=potentialMergeMap.end())
+		{
+			int newEndCol=potentialMergeMap.find(col)->second;
+			std::map<Vec2i, LotusSpreadsheetInternal::Cell>::const_iterator newCIt=cIt;
+			if (newCIt!=sheet.m_positionToCellMap.end()) ++newCIt;
+			canMerge=(newCIt==sheet.m_positionToCellMap.end() || newCIt->first[1]!=row || newCIt->first[0]>=newEndCol);
+		}
+		if (canMerge)
+		{
+			int newEndCol=potentialMergeMap.find(col)->second;
+			while (colIt!=newColSet.end() && *colIt<newEndCol)
+			{
+				++colIt;
+				continue;
+			}
+			LotusSpreadsheetInternal::Cell cell=cIt->second;
+			cell.setNumSpannedCells(Vec2i(newEndCol-col,1));
+			sendCellContent(cell, style, 1);
+		}
+		else
+			sendCellContent(hasCell ? cIt->second : emptyCell, style, endCol-col);
 	}
 }
 
@@ -2005,7 +2728,9 @@ void LotusSpreadsheet::sendCellContent(LotusSpreadsheetInternal::Cell const &cel
 	}
 	if (cell.m_comment.valid())
 	{
-		WPS_DEBUG_MSG(("LotusSpreadsheet::sendCellContent: oops sending comment is not implemented\n"));
+		WPSSubDocumentPtr subdoc(new LotusSpreadsheetInternal::SubDocument
+		                         (cell.m_input, *this, cell.m_comment));
+		m_listener->insertComment(subdoc);
 	}
 	m_listener->closeSheetCell();
 }
@@ -2095,7 +2820,7 @@ static Functions const s_listFunctions[] =
 	{ "DVar", 3},{ "DStd", 3},{ "Index", 3}, { "Columns", 1},
 	{ "Rows", 1},{ "Rept", 2},{ "Upper", 1},{ "Lower", 1},
 	{ "Left", 2},{ "Right", 2},{ "Replace", 4}, { "Proper", 1},
-	{ "Cell", 1} /*checkme*/,{ "Trim", 1},{ "Clean", 1} /*UNKN*/,{ "T", 1},
+	{ "Cell", 2} /*checkme*/,{ "Trim", 1},{ "Clean", 1} /*UNKN*/,{ "T", 1},
 
 	{ "IsNonText", 1},{ "Exact", 2},{ "", -2} /*App not implemented*/,{ "", 3} /*UNKN*/,
 	{ "Rate", 3} /*BAD*/,{ "TERM", 3}, { "CTERM", 3}, { "SLN", 3},
@@ -2440,6 +3165,7 @@ bool LotusSpreadsheet::readFormula(WPSStream &stream, long endPos, int sheetId,
 	f.str("");
 	for (size_t i = 0; i < stack.size(); ++i)
 	{
+		if (i) f << "##";
 		for (size_t j=0; j < stack[i].size(); ++j)
 			f << stack[i][j] << ",";
 	}
@@ -2629,6 +3355,43 @@ void LotusSpreadsheet::sendText(RVNGInputStreamPtr &input, long endPos, LotusSpr
 	}
 }
 
+void LotusSpreadsheet::sendTextNote(RVNGInputStreamPtr &input, WPSEntry const &entry) const
+{
+	if (!input || !m_listener)
+	{
+		WPS_DEBUG_MSG(("LotusSpreadsheet::sendTextNote: can not find the listener\n"));
+		return;
+	}
+	bool prevEOL=false;
+	libwps_tools_win::Font::Type fontType=m_mainParser.getDefaultFontType();
+	input->seek(entry.begin(), librevenge::RVNG_SEEK_SET);
+	long endPos=entry.end();
+	while (!input->isEnd())
+	{
+		long pos=input->tell();
+		if (pos>=endPos) break;
+		char c=char(libwps::readU8(input));
+		switch (c)
+		{
+		case 0xd:
+			m_listener->insertEOL();
+			prevEOL=true;
+			break;
+		case 0xa:
+			if (!prevEOL)
+			{
+				WPS_DEBUG_MSG(("LotusSpreadsheet::sendTextNote: find 0xa without 0xd\n"));
+			}
+			prevEOL=false;
+			break;
+		default:
+			if (c)
+				m_listener->insertUnicode((uint32_t)libwps_tools_win::Font::unicode(static_cast<unsigned char>(c),fontType));
+			break;
+		}
+	}
+}
+
 std::string LotusSpreadsheet::getDebugStringForText(std::string const &text)
 {
 	std::string res;
