@@ -36,6 +36,7 @@
 #include "WKSContentListener.h"
 #include "WPSEntry.h"
 #include "WPSFont.h"
+#include "WPSTable.h"
 
 #include "Quattro.h"
 
@@ -49,7 +50,7 @@ namespace QuattroSpreadsheetInternal
 struct Style : public WPSCellFormat
 {
 	//! construtor
-	Style(libwps_tools_win::Font::Type type) : WPSCellFormat(), m_fontType(type), m_fileFormat(0xFF), m_extra("")
+	explicit Style(libwps_tools_win::Font::Type type) : WPSCellFormat(), m_fontType(type), m_fileFormat(0xFF), m_extra("")
 	{
 	}
 
@@ -139,7 +140,7 @@ class Cell : public WPSCell
 {
 public:
 	/// constructor
-	Cell(libwps_tools_win::Font::Type type) : m_fontType(type), m_fileFormat(0xFF), m_content() { }
+	explicit Cell(libwps_tools_win::Font::Type type) : m_fontType(type), m_fileFormat(0xFF), m_content() { }
 
 	//! operator<<
 	friend std::ostream &operator<<(std::ostream &o, Cell const &cell);
@@ -277,30 +278,13 @@ std::ostream &operator<<(std::ostream &o, Cell const &cell)
 //! the spreadsheet of a WPS4Spreadsheet
 class Spreadsheet
 {
-protected:
-	//! a comparaison structure used to sort cell by rows and and columns
-	struct ComparePosition
-	{
-		//! constructor
-		ComparePosition() {}
-		//! comparaison function
-		bool operator()(Vec2i const &c1, Vec2i const &c2) const
-		{
-			if (c1[1]!=c2[1])
-				return c1[1]<c2[1];
-			return c1[0]<c2[0];
-		}
-	};
-
 public:
 	//! the spreadsheet type
 	enum Type { T_Spreadsheet, T_Filter, T_Report };
 
-	typedef std::map<Vec2i, Cell, ComparePosition> PositionToCellMap_t;
-
 	//! a constructor
-	Spreadsheet(Type type=T_Spreadsheet, int id=0) : m_type(type), m_id(id), m_numCols(0), m_numRows(0), m_LBPosition(-1,-1),
-		m_widthCols(), m_heightRows(), m_positionToCellMap(), m_lastCellPos(),
+	Spreadsheet(Type type=T_Spreadsheet, int id=0) : m_type(type), m_id(id), m_numCols(0), m_numRows(0),
+		m_widthCols(), m_rowHeightMap(), m_heightDefault(16), m_positionToCellMap(), m_lastCellPos(),
 		m_rowPageBreaksList() {}
 	//! return a cell corresponding to a spreadsheet, create one if needed
 	Cell &getCell(Vec2i const &pos, libwps_tools_win::Font::Type type)
@@ -309,8 +293,7 @@ public:
 		{
 			Cell cell(type);
 			cell.setPosition(pos);
-			m_positionToCellMap.insert
-			(std::map<Vec2i, Cell, ComparePosition>::value_type(pos,cell));
+			m_positionToCellMap.insert(std::map<Vec2i, Cell>::value_type(pos,cell));
 		}
 		m_lastCellPos=pos;
 		return m_positionToCellMap.find(pos)->second;
@@ -330,40 +313,99 @@ public:
 		m_widthCols[size_t(col)] = w;
 		if (col >= m_numCols) m_numCols=col+1;
 	}
-	//! set the rows size
-	void setRowHeight(int row, int h=-1)
-	{
-		if (row < 0) return;
-		if (row >= int(m_heightRows.size())) m_heightRows.resize(size_t(row)+1, -1);
-		m_heightRows[size_t(row)] = h;
-		if (row >= m_numRows) m_numRows=row+1;
-	}
 
-	//! convert the m_widthCols, m_heightRows in a vector of of point size
-	static std::vector<float> convertInPoint(std::vector<int> const &list,
-	                                         float defSize)
+	//! return the columns format
+	std::vector<WPSColumnFormat> getWidths(float defSize=76) const
 	{
-		size_t numElt = list.size();
-		std::vector<float> res;
-		res.resize(numElt);
-		for (size_t i = 0; i < numElt; i++)
+		std::vector<WPSColumnFormat> widths;
+		WPSColumnFormat defWidth(defSize), actWidth;
+		defWidth.m_useOptimalWidth=true;
+		int repeat=0;
+		for (size_t c = 0; c < m_widthCols.size(); c++)
 		{
-			if (list[i] < 0) res[i] = defSize;
-			else res[i] = float(list[i])/20.f;
+			WPSColumnFormat newWidth;
+			if (m_widthCols[c] < 0)
+				newWidth=defWidth;
+			else
+				newWidth=WPSColumnFormat(float(m_widthCols[c])/20.f);
+			if (repeat && newWidth!=actWidth)
+			{
+				actWidth.m_numRepeat=repeat;
+				widths.push_back(actWidth);
+				repeat=0;
+			}
+			if (repeat==0)
+				actWidth=newWidth;
+			++repeat;
 		}
-		return res;
+		if (repeat)
+		{
+			actWidth.m_numRepeat=repeat;
+			widths.push_back(actWidth);
+		}
+		return widths;
 	}
-	/** compute the last Right Bottom cell position */
-	void computeRightBottomPosition()
+	//! returns the row size in point
+	float getRowHeight(int row) const
 	{
-		int maxX = -1, maxY = -1;
-		for (PositionToCellMap_t::const_iterator it=m_positionToCellMap.begin(); it!=m_positionToCellMap.end(); ++it)
+		std::map<Vec2i,int>::const_iterator rIt=m_rowHeightMap.lower_bound(Vec2i(-1,row));
+		if (rIt!=m_rowHeightMap.end() && rIt->first[0]<=row && rIt->first[1]>=row)
+			return (float) rIt->second;
+		return (float) m_heightDefault;
+	}
+	//! returns the height of a row in point and updated repeated row
+	float getRowHeight(int row, int &numRepeated) const
+	{
+		std::map<Vec2i,int>::const_iterator rIt=m_rowHeightMap.lower_bound(Vec2i(-1,row));
+		if (rIt!=m_rowHeightMap.end() && rIt->first[0]<=row && rIt->first[1]>=row)
 		{
-			Vec2i const &p = it->second.position();
-			if (p[0] > maxX) maxX = p[0];
-			if (p[1] > maxY) maxY = p[1];
+			numRepeated=rIt->first[1]-row+1;
+			return (float) rIt->second;
 		}
-		m_LBPosition=Vec2i(maxX, maxY);
+		numRepeated=10000;
+		return (float) m_heightDefault;
+	}
+	//! set the rows size
+	void setRowHeight(int row, int h)
+	{
+		if (h>=0)
+			m_rowHeightMap[Vec2i(row,row)]=h;
+	}
+	//! try to compress the list of row height
+	void compressRowHeights()
+	{
+		std::map<Vec2i,int> oldMap=m_rowHeightMap;
+		m_rowHeightMap.clear();
+		std::map<Vec2i,int>::const_iterator rIt=oldMap.begin();
+		int actHeight=-1;
+		Vec2i actPos(0,-1);
+		while (rIt!=oldMap.end())
+		{
+			// first check for not filled row
+			if (rIt->first[0]!=actPos[1]+1)
+			{
+				if (actHeight==m_heightDefault)
+					actPos[1]=rIt->first[0]-1;
+				else
+				{
+					if (actPos[1]>=actPos[0])
+						m_rowHeightMap[actPos]=actHeight;
+					actHeight=m_heightDefault;
+					actPos=Vec2i(actPos[1]+1, rIt->first[0]-1);
+				}
+			}
+			if (rIt->second!=actHeight)
+			{
+				if (actPos[1]>=actPos[0])
+					m_rowHeightMap[actPos]=actHeight;
+				actPos[0]=rIt->first[0];
+				actHeight=rIt->second;
+			}
+			actPos[1]=rIt->first[1];
+			++rIt;
+		}
+		if (actPos[1]>=actPos[0])
+			m_rowHeightMap[actPos]=actHeight;
 	}
 	//! returns true if the spreedsheet is empty
 	bool empty() const
@@ -378,15 +420,15 @@ public:
 	int m_numCols;
 	/** the number of rows */
 	int m_numRows;
-	/** the final Right Bottom position, computed by updateState */
-	Vec2i m_LBPosition;
 
 	/** the column size in TWIP (?) */
 	std::vector<int> m_widthCols;
-	/** the row size in TWIP (?) */
-	std::vector<int> m_heightRows;
+	/** the map Vec2i(min row, max row) to size in points */
+	std::map<Vec2i,int> m_rowHeightMap;
+	/** the default row size in point */
+	int m_heightDefault;
 	/** a map cell to not empty cells */
-	PositionToCellMap_t m_positionToCellMap;
+	std::map<Vec2i, Cell> m_positionToCellMap;
 	/** the last cell position */
 	Vec2i m_lastCellPos;
 	/** the list of row page break */
@@ -398,7 +440,7 @@ public:
 struct State
 {
 	//! constructor
-	State() :  m_eof(-1), m_version(-1), m_hasLICSCharacters(-1), m_styleManager(), m_spreadsheetList(), m_spreadsheetStack()
+	State() :  m_eof(-1), m_version(-1), m_maxDimension(0,0,0), m_hasLICSCharacters(-1), m_styleManager(), m_spreadsheetList(), m_spreadsheetStack()
 	{
 		pushNewSheet(shared_ptr<Spreadsheet>(new Spreadsheet(Spreadsheet::T_Spreadsheet, 0)));
 	}
@@ -464,6 +506,8 @@ struct State
 	long m_eof;
 	//! the file version
 	int m_version;
+	//! the maximum col, row, sheet
+	WPSVec3i m_maxDimension;
 	//! int to code if the file has LICS characters:-1 means unknown, 0 means no, 1 means yes
 	int m_hasLICSCharacters;
 	//! the style manager
@@ -537,13 +581,14 @@ bool QuattroSpreadsheet::readSheetSize()
 		return false;
 	}
 	long sz = libwps::readU16(m_input);
-	if (sz < 8)
+	int const vers=version();
+	if (sz < (vers>=2 ? 12 : 8))
 	{
 		WPS_DEBUG_MSG(("QuattroSpreadsheet::readSheetSize: block is too short\n"));
 		return false;
 	}
 	f << "Entries(SheetSize):";
-	for (int i = 0; i < 2; i++)   // always 2 zero ?
+	for (int i = 0; i < (vers>=2 ? 3 : 2); i++)   // always 2 zero ?
 	{
 		int val = libwps::read16(m_input);
 		if (!val) continue;
@@ -553,13 +598,15 @@ bool QuattroSpreadsheet::readSheetSize()
 	f << "nCols=" << nCol << ",";
 	int nRow =  libwps::read16(m_input);
 	f << "nRow=" << nRow << ",";
+	int nSheet =  vers<=1 ? 0 : libwps::read16(m_input);
+	if (nSheet>0)
+		f << "nSheet=" << nSheet << ",";
+	m_state->m_maxDimension=WPSVec3i(nCol,nRow,nSheet);
 	ascii().addPos(pos);
 	ascii().addNote(f.str().c_str());
 	// empty spreadsheet
 	if (nRow==-1 && nCol==0) return true;
 	if (nRow < 0 || nCol <= 0) return false;
-
-	m_state->getActualSheet().setRowHeight(nRow-1);
 	m_state->getActualSheet().setColumnWidth(nCol-1);
 	return true;
 
@@ -585,25 +632,13 @@ bool QuattroSpreadsheet::readRowSize()
 	int row = libwps::read16(m_input);
 	int height = libwps::readU8(m_input);
 
-	bool ok = row >= 0 && row < m_state->getActualSheet().m_numRows+20;
+	bool ok = row >= 0;
 	f << "Entries(Row):Row" << row << "";
 	if (!ok) f << "###";
 	f << ":height=" << height << ",";
 
 	if (ok)
-	{
-		if (row >= m_state->getActualSheet().m_numRows)
-		{
-			static bool first = true;
-			if (first)
-			{
-				first = false;
-				WPS_DEBUG_MSG(("QuattroSpreadsheet::readRowSize: I must increase the number of rows\n"));
-			}
-			f << "#row[inc],";
-		}
-		m_state->getActualSheet().setRowHeight(row, 20*height); // point to TWIP
-	}
+		m_state->getActualSheet().setRowHeight(row, height); // point to TWIP
 	ascii().addPos(pos);
 	ascii().addNote(f.str().c_str());
 
@@ -1267,9 +1302,11 @@ bool QuattroSpreadsheet::readUserStyle()
 		break;
 	}
 	style.m_fileFormat=(int) libwps::readU8(m_input);
-	int dim[2];
-	for (int i=0; i<2; ++i) dim[i]=(int) libwps::read16(m_input);
-	if (dim[0]||dim[1]) f << "dim=" << Vec2i(dim[0],dim[1]) << ",";
+	for (int i=0; i<3; ++i)
+	{
+		val=(int) libwps::read8(m_input);
+		if (val) f << "f" << i+2 << "=" << val << ",";
+	}
 
 	style.m_extra=f.str();
 	m_state->m_styleManager.add(id, style);
@@ -1783,7 +1820,6 @@ bool QuattroSpreadsheet::readFormula(long endPos, Vec2i const &position, int she
 	if (endPos-pos-2 != sz) return false;
 
 	std::vector<WKSContentListener::FormulaInstruction> listCellsPos;
-	size_t actCellId=0;
 	int fieldPos[2]= {0,sz};
 	for (int i=0; i<2; ++i)
 	{
@@ -1796,6 +1832,7 @@ bool QuattroSpreadsheet::readFormula(long endPos, Vec2i const &position, int she
 		error="###fieldPos";
 		return false;
 	}
+	int NSingle=0;
 	if (fieldPos[0]!=sz)
 	{
 		m_input->seek(pos+2+fieldPos[0], librevenge::RVNG_SEEK_SET);
@@ -1803,6 +1840,7 @@ bool QuattroSpreadsheet::readFormula(long endPos, Vec2i const &position, int she
 		if (vers==1)
 		{
 			int N=(int)(sz-fieldPos[0])/4;
+			NSingle=(int)(fieldPos[1]-fieldPos[0])/4;
 			for (int i=0; i<N; ++i)
 			{
 				WKSContentListener::FormulaInstruction cell;
@@ -1858,6 +1896,7 @@ bool QuattroSpreadsheet::readFormula(long endPos, Vec2i const &position, int she
 	std::stringstream f;
 	std::vector<std::vector<WKSContentListener::FormulaInstruction> > stack;
 	bool ok = true;
+	size_t actCellId=0, actDualCellId=size_t(NSingle);
 	while (long(m_input->tell()) != endPos)
 	{
 		double val;
@@ -1906,19 +1945,19 @@ bool QuattroSpreadsheet::readFormula(long endPos, Vec2i const &position, int she
 				instr=listCellsPos[actCellId++];
 				break;
 			}
-			if (actCellId+1>=listCellsPos.size())
+			if (actDualCellId+1>=listCellsPos.size())
 			{
 				f.str("");
-				f << "###unknListCell" << actCellId;
+				f << "###unknListCell" << actDualCellId;
 				error=f.str();
 				ok = false;
 				break;
 			}
-			instr=listCellsPos[actCellId++];
+			instr=listCellsPos[actDualCellId++];
 			instr.m_type=WKSContentListener::FormulaInstruction::F_CellList;
-			instr.m_position[1]=listCellsPos[actCellId].m_position[0];
-			instr.m_positionRelative[1]=listCellsPos[actCellId].m_positionRelative[0];
-			++actCellId;
+			instr.m_position[1]=listCellsPos[actDualCellId].m_position[0];
+			instr.m_positionRelative[1]=listCellsPos[actDualCellId].m_positionRelative[0];
+			++actDualCellId;
 			break;
 		case 0x5:
 			instr.m_type=WKSContentListener::FormulaInstruction::F_Long;
@@ -2146,30 +2185,36 @@ void QuattroSpreadsheet::sendSpreadsheet(int sId)
 		}
 		sheet.reset(new QuattroSpreadsheetInternal::Spreadsheet);
 	}
-	sheet->computeRightBottomPosition();
 
-	m_listener->openSheet(sheet->convertInPoint(sheet->m_widthCols,76), librevenge::RVNG_POINT,
-	                      m_state->getSheetName(sId));
-	std::vector<float> rowHeight = sheet->convertInPoint(sheet->m_heightRows,16);
-
-	typedef std::map<int, const QuattroSpreadsheetInternal::Cell *> SparseRow_t;
-	typedef std::map<int, SparseRow_t> SparseTable_t;
-	SparseTable_t table;
-	for (QuattroSpreadsheetInternal::Spreadsheet::PositionToCellMap_t::const_iterator it = sheet->m_positionToCellMap.begin(); it != sheet->m_positionToCellMap.end(); ++it)
-		table[it->first[1]][it->first[0]] = &it->second;
-
-	SparseTable_t::const_iterator rIt = table.begin();
-	for (int r=0; r<=sheet->m_LBPosition[1] && rIt != table.end(); ++r)
+	m_listener->openSheet(sheet->getWidths(), m_state->getSheetName(sId));
+	sheet->compressRowHeights();
+	std::map<Vec2i, QuattroSpreadsheetInternal::Cell>::const_iterator it = sheet->m_positionToCellMap.begin();
+	int prevRow = -1;
+	while (it != sheet->m_positionToCellMap.end())
 	{
-		m_listener->openSheetRow(r < int(rowHeight.size()) ? rowHeight[size_t(r)] : 16, librevenge::RVNG_POINT);
-		if (r == rIt->first) // a row with at least one non-empty cell
+		int row=it->first[1];
+		QuattroSpreadsheetInternal::Cell const &cell=(it++)->second;
+		if (row>prevRow+1)
 		{
-			for (SparseRow_t::const_iterator cIt = rIt->second.begin(); cIt != rIt->second.end(); ++cIt)
-				sendCellContent(*cIt->second);
-			++rIt;
+			while (row > prevRow+1)
+			{
+				if (prevRow != -1) m_listener->closeSheetRow();
+				int numRepeat;
+				float h=sheet->getRowHeight(prevRow+1, numRepeat);
+				if (row<prevRow+1+numRepeat)
+					numRepeat=row-1-prevRow;
+				m_listener->openSheetRow(WPSRowFormat(h), numRepeat);
+				prevRow+=numRepeat;
+			}
 		}
-		m_listener->closeSheetRow();
+		if (row!=prevRow)
+		{
+			if (prevRow != -1) m_listener->closeSheetRow();
+			m_listener->openSheetRow(WPSRowFormat(sheet->getRowHeight(++prevRow)));
+		}
+		sendCellContent(cell);
 	}
+	if (prevRow!=-1) m_listener->closeSheetRow();
 	m_listener->closeSheet();
 }
 
@@ -2181,7 +2226,6 @@ void QuattroSpreadsheet::sendCellContent(QuattroSpreadsheetInternal::Cell const
 		return;
 	}
 
-	librevenge::RVNGPropertyList propList;
 	libwps_tools_win::Font::Type fontType = cell.m_fontType;
 	m_listener->setFont(cell.getFont());
 
@@ -2197,16 +2241,16 @@ void QuattroSpreadsheet::sendCellContent(QuattroSpreadsheetInternal::Cell const
 		for (size_t c=0; c < text.length(); ++c)
 		{
 			if (!hasLICS)
-				WPSListener::appendUnicode
+				libwps::appendUnicode
 				((uint32_t)libwps_tools_win::Font::unicode((unsigned char)text[c],fontType), finalString);
 			else
-				WPSListener::appendUnicode
-				((uint32_t)libwps_tools_win::Font::LICSunicode((unsigned char)text[c],fontType), finalString);
+				libwps::appendUnicode
+				((uint32_t)libwps_tools_win::Font::LICSunicode((unsigned char)text[c]), finalString);
 		}
 		text=finalString.cstr();
 	}
 	finalCell.updateFormat();
-	m_listener->openSheetCell(finalCell, content, propList);
+	m_listener->openSheetCell(finalCell, content);
 
 	if (cell.m_content.m_textEntry.valid())
 	{
@@ -2233,7 +2277,7 @@ void QuattroSpreadsheet::sendCellContent(QuattroSpreadsheetInternal::Cell const
 				if (!hasLICS)
 					m_listener->insertUnicode((uint32_t)libwps_tools_win::Font::unicode(c,fontType));
 				else
-					m_listener->insertUnicode((uint32_t)libwps_tools_win::Font::LICSunicode(c,fontType));
+					m_listener->insertUnicode((uint32_t)libwps_tools_win::Font::LICSunicode(c));
 				prevEOL=false;
 			}
 		}
