@@ -36,6 +36,7 @@
 #include "WKSContentListener.h"
 #include "WPSEntry.h"
 #include "WPSFont.h"
+#include "WPSTable.h"
 
 #include "WKS4.h"
 
@@ -337,19 +338,36 @@ public:
 		if (actPos[1]>=actPos[0])
 			m_rowHeightMap[actPos]=actHeight;
 	}
-	//! convert the m_widthCols in a vector of of point size
-	static std::vector<float> convertInPoint(std::vector<int> const &list,
-	                                         float defSize)
+	//! return the columns format
+	std::vector<WPSColumnFormat> getWidths(float defSize=72) const
 	{
-		size_t numElt = list.size();
-		std::vector<float> res;
-		res.resize(numElt);
-		for (size_t i = 0; i < numElt; i++)
+		std::vector<WPSColumnFormat> widths;
+		WPSColumnFormat defWidth(defSize), actWidth;
+		defWidth.m_useOptimalWidth=true;
+		int repeat=0;
+		for (size_t c = 0; c < m_widthCols.size(); c++)
 		{
-			if (list[i] < 0) res[i] = defSize;
-			else res[i] = float(list[i])/20.f;
+			WPSColumnFormat newWidth;
+			if (m_widthCols[c] < 0)
+				newWidth=defWidth;
+			else
+				newWidth=WPSColumnFormat(float(m_widthCols[c])/20.f);
+			if (repeat && newWidth!=actWidth)
+			{
+				actWidth.m_numRepeat=repeat;
+				widths.push_back(actWidth);
+				repeat=0;
+			}
+			if (repeat==0)
+				actWidth=newWidth;
+			++repeat;
 		}
-		return res;
+		if (repeat)
+		{
+			actWidth.m_numRepeat=repeat;
+			widths.push_back(actWidth);
+		}
+		return widths;
 	}
 	//! returns true if the spreedsheet is empty
 	bool empty() const
@@ -475,6 +493,11 @@ WKS4Spreadsheet::~WKS4Spreadsheet()
 {
 }
 
+void WKS4Spreadsheet::resetInput(RVNGInputStreamPtr newInput)
+{
+	m_input=newInput;
+}
+
 int WKS4Spreadsheet::version() const
 {
 	if (m_state->m_version<0)
@@ -798,7 +821,7 @@ bool WKS4Spreadsheet::readMsWorksStyle()
 			break;
 		case 5: // center arround cell
 			f << ",center[between cell]";
-		// fail through expected
+			WPS_FALLTHROUGH;
 		case 2:
 			style.setHAlignement(WPSCell::HALIGN_CENTER);
 			break;
@@ -1832,7 +1855,7 @@ static Functions const s_listFunctions[] =
 
 	{ "Time", 3},{ "Hour", 1},{ "Minute", 1},{ "Second", 1},
 	{ "IsNumber", 1},{ "IsText", 1},{ "Len", 1},{ "Value", 1},
-	{ "Text", 2}/* or fixed*/, { "Mid", 3}, { "Char", 1},{ "Ascii", 1},
+	{ "Fixed", 2}, { "Mid", 3}, { "Char", 1},{ "Ascii", 1},
 	{ "Find", 3},{ "DateValue", 1} /*checkme*/,{ "TimeValue", 1} /*checkme*/,{ "CellPointer", 1} /*checkme*/,
 
 	{ "Sum", -1},{ "Average", -1},{ "COUNT", -1},{ "Min", -1},
@@ -2322,8 +2345,7 @@ void WKS4Spreadsheet::sendSpreadsheet(int sId)
 		sheet.reset(new WKS4SpreadsheetInternal::Spreadsheet);
 	}
 
-	m_listener->openSheet(sheet->convertInPoint(sheet->m_widthCols,72), librevenge::RVNG_POINT,
-	                      std::vector<int>(), m_state->getSheetName(sId));
+	m_listener->openSheet(sheet->getWidths(), m_state->getSheetName(sId));
 	sheet->compressRowHeights();
 	std::map<Vec2i, WKS4SpreadsheetInternal::Cell>::const_iterator it = sheet->m_positionToCellMap.begin();
 	int prevRow = -1;
@@ -2340,14 +2362,14 @@ void WKS4Spreadsheet::sendSpreadsheet(int sId)
 				float h=sheet->getRowHeight(prevRow+1, numRepeat);
 				if (row<prevRow+1+numRepeat)
 					numRepeat=row-1-prevRow;
-				m_listener->openSheetRow(h, librevenge::RVNG_POINT, false, numRepeat);
+				m_listener->openSheetRow(WPSRowFormat(h), numRepeat);
 				prevRow+=numRepeat;
 			}
 		}
 		if (row!=prevRow)
 		{
 			if (prevRow != -1) m_listener->closeSheetRow();
-			m_listener->openSheetRow(sheet->getRowHeight(++prevRow), librevenge::RVNG_POINT);
+			m_listener->openSheetRow(WPSRowFormat(sheet->getRowHeight(++prevRow)));
 		}
 		sendCellContent(cell);
 	}
@@ -2392,7 +2414,7 @@ void WKS4Spreadsheet::sendCellContent(WKS4SpreadsheetInternal::Cell const &cell)
 				((uint32_t)libwps_tools_win::Font::unicode((unsigned char)text[c],fontType), finalString);
 			else
 				libwps::appendUnicode
-				((uint32_t)libwps_tools_win::Font::LICSunicode((unsigned char)text[c],fontType), finalString);
+				((uint32_t)libwps_tools_win::Font::LICSunicode((unsigned char)text[c]), finalString);
 		}
 		text=finalString.cstr();
 	}
@@ -2428,7 +2450,7 @@ void WKS4Spreadsheet::sendCellContent(WKS4SpreadsheetInternal::Cell const &cell)
 				if (!hasLICS)
 					m_listener->insertUnicode((uint32_t)libwps_tools_win::Font::unicode(c,fontType));
 				else
-					m_listener->insertUnicode((uint32_t)libwps_tools_win::Font::LICSunicode(c,fontType));
+					m_listener->insertUnicode((uint32_t)libwps_tools_win::Font::LICSunicode(c));
 				prevEOL=false;
 			}
 		}
