@@ -40,6 +40,7 @@
 #include "WPSPageSpan.h"
 #include "WPSParagraph.h"
 #include "WPSPosition.h"
+#include "WPSTable.h"
 #include "WKSSubDocument.h"
 
 ////////////////////////////////////////////////////////////
@@ -102,6 +103,8 @@ struct WKSContentParsingState
 
 	WPSFont m_font;
 	WPSParagraph m_paragraph;
+	//! a flag to know if openGroup was called
+	bool m_isGroupOpened;
 
 	bool m_isParagraphColumnBreak;
 	bool m_isParagraphPageBreak;
@@ -132,7 +135,7 @@ WKSContentParsingState::WKSContentParsingState() :
 	m_textBuffer(""), m_numDeferredTabs(0),
 
 	m_font(), m_paragraph(),
-
+	m_isGroupOpened(false),
 	m_isParagraphColumnBreak(false), m_isParagraphPageBreak(false),
 
 	m_isSpanOpened(false), m_isParagraphOpened(false),
@@ -526,13 +529,18 @@ void WKSContentListener::insertComment(WPSSubDocumentPtr &subDocument)
 		return;
 	}
 
-	if (!m_ps->m_isParagraphOpened)
-		_openParagraph();
-	else
+	if (!m_ps->m_isSheetCellOpened)
 	{
-		_flushText();
-		_closeSpan();
+		if (!m_ps->m_isParagraphOpened)
+			_openParagraph();
+		else
+		{
+			_flushText();
+			_closeSpan();
+		}
 	}
+	else if (m_ps->m_isParagraphOpened)
+		_closeParagraph();
 
 	librevenge::RVNGPropertyList propList;
 	m_documentInterface->openComment(propList);
@@ -628,6 +636,42 @@ void WKSContentListener::insertPicture
 ///////////////////
 // frame
 ///////////////////
+bool WKSContentListener::openGroup(WPSPosition const &pos)
+{
+	if (!m_ds->m_isDocumentStarted)
+	{
+		WPS_DEBUG_MSG(("WKSContentListener::openGroup: the document is not started\n"));
+		return false;
+	}
+	if (m_ps->m_isSheetRowOpened)
+	{
+		WPS_DEBUG_MSG(("WKSContentListener::openGroup: can not open a group\n"));
+		return false;
+	}
+	librevenge::RVNGPropertyList propList;
+	_handleFrameParameters(propList, pos);
+
+	_pushParsingState();
+	_startSubDocument();
+	m_ps->m_isGroupOpened = true;
+
+	m_documentInterface->openGroup(propList);
+
+	return true;
+}
+
+void WKSContentListener::closeGroup()
+{
+	if (!m_ps->m_isGroupOpened)
+	{
+		WPS_DEBUG_MSG(("WKSContentListener::closeGroup: called but no group is already opened\n"));
+		return;
+	}
+	_endSubDocument();
+	_popParsingState();
+	m_documentInterface->closeGroup();
+}
+
 bool WKSContentListener::_openFrame(WPSPosition const &pos, WPSGraphicStyle const &style)
 {
 	if (m_ps->m_isFrameOpened)
@@ -731,7 +775,7 @@ void WKSContentListener::_handleFrameParameters
 	{
 	case WPSPosition::YFull:
 		propList.insert("svg:height", double(h), unit);
-	// fall-through intended
+		WPS_FALLTHROUGH;
 	case WPSPosition::YTop:
 		if (origin[1] < 0.0 || origin[1] > 0.0)
 		{
@@ -775,7 +819,7 @@ void WKSContentListener::_handleFrameParameters
 	{
 	case WPSPosition::XFull:
 		propList.insert("svg:width", double(w), unit);
-	// fallthrough intended
+		WPS_FALLTHROUGH;
 	case WPSPosition::XLeft:
 		if (origin[0] < 0.0 || origin[0] > 0.0)
 		{
@@ -876,8 +920,7 @@ void WKSContentListener::_endSubDocument()
 ///////////////////
 // sheet
 ///////////////////
-void WKSContentListener::openSheet(std::vector<float> const &colWidth, librevenge::RVNGUnit unit,
-                                   std::vector<int> const &repeatColWidthNumber, librevenge::RVNGString const &name)
+void WKSContentListener::openSheet(std::vector<WPSColumnFormat> const &colList, librevenge::RVNGString const &name)
 {
 	if (m_ps->m_isSheetOpened)
 	{
@@ -898,18 +941,10 @@ void WKSContentListener::openSheet(std::vector<float> const &colWidth, libreveng
 
 	librevenge::RVNGPropertyListVector columns;
 
-	size_t nCols = colWidth.size();
-	bool useRepeated=repeatColWidthNumber.size()==nCols;
-	if (!useRepeated&&!repeatColWidthNumber.empty())
-	{
-		WPS_DEBUG_MSG(("WKSContentListener::openSheet: repeatColWidthNumber seems bad\n"));
-	}
-	for (size_t c = 0; c < nCols; c++)
+	for (size_t c = 0; c < colList.size(); c++)
 	{
 		librevenge::RVNGPropertyList column;
-		column.insert("style:column-width", colWidth[c], unit);
-		if (useRepeated && repeatColWidthNumber[c]>1)
-			column.insert("table:number-columns-repeated", repeatColWidthNumber[c]);
+		colList[c].addTo(column);
 		columns.append(column);
 	}
 	propList.insert("librevenge:columns", columns);
@@ -934,7 +969,7 @@ void WKSContentListener::closeSheet()
 	_popParsingState();
 }
 
-void WKSContentListener::openSheetRow(float h, librevenge::RVNGUnit unit, bool headerRow, int numRepeated)
+void WKSContentListener::openSheetRow(WPSRowFormat const &format, int numRepeated)
 {
 	if (m_ps->m_isSheetRowOpened)
 	{
@@ -947,14 +982,9 @@ void WKSContentListener::openSheetRow(float h, librevenge::RVNGUnit unit, bool h
 		return;
 	}
 	librevenge::RVNGPropertyList propList;
-	propList.insert("librevenge:is-header-row", headerRow);
+	format.addTo(propList);
 	if (numRepeated>1)
 		propList.insert("table:number-rows-repeated", numRepeated);
-
-	if (h > 0)
-		propList.insert("style:row-height", h, unit);
-	else if (h < 0)
-		propList.insert("style:min-row-height", -h, unit);
 	m_documentInterface->openSheetRow(propList);
 	m_ps->m_isSheetRowOpened = true;
 }
@@ -1185,6 +1215,7 @@ shared_ptr<WKSContentParsingState> WKSContentListener::_pushParsingState()
 	m_ps->m_pageMarginBottom = actual->m_pageMarginBottom;
 
 	m_ps->m_isNote = actual->m_isNote;
+	m_ps->m_isPageSpanOpened = actual->m_isPageSpanOpened;
 
 	return actual;
 }
