@@ -585,7 +585,7 @@ bool WPS4Parser::findZones()
 		{
 		case 0xda1:
 			apCreator = 2;
-		// fall-through intended
+			WPS_FALLTHROUGH;
 		case 0:
 		case 1:
 			worksVersion = 1;
@@ -752,7 +752,7 @@ bool WPS4Parser::findZones()
 	}
 
 	int num = libwps::readU16(input);
-	int sz = libwps::readU16(input);
+	long sz = libwps::readU16(input);
 	begP = (long) libwps::readU32(input);
 	if (begP)
 	{
