@@ -910,6 +910,7 @@ bool WPS4Text::readText(WPSEntry const &zone)
 					}
 					m_input->seek(-1, librevenge::RVNG_SEEK_CUR);
 				}
+				WPS_FALLTHROUGH;
 			default:
 				if (version()<=2)
 				{
@@ -1294,7 +1295,7 @@ bool WPS4Text::findFDPStructuresByHand(int which)
 }
 
 // PLC Data: default parser
-bool WPS4Text::defDataParser(long , long , int , long endPos, std::string &mess)
+bool WPS4Text::defDataParser(long, long, int, long endPos, std::string &mess)
 {
 	mess = "";
 	libwps::DebugStream f;
@@ -1593,7 +1594,7 @@ bool WPS4Text::readDosLink(WPSEntry const &entry)
 			val = libwps::readU16(m_input); // always 4
 			if (val != 4) f << "g1=" << val << ",";
 		}
-		// fall-through intended
+		WPS_FALLTHROUGH;
 		case 0x40: // spreadsheet range
 		case 0x01: // char ?
 		{
@@ -1999,7 +2000,7 @@ bool WPS4Text::readParagraph(long endPos, int &id, std::string &mess)
 				break;
 			case 0x13: // seems another way to define the left margin
 				f << "#left,";
-			// fall-through intended
+				WPS_FALLTHROUGH;
 			case 0x12:
 				pp.m_margins[1] = arg/1440.;
 				break;
@@ -2306,7 +2307,9 @@ bool WPS4Text::objectDataParser(long bot, long /*eot*/, int id,
 		if (val) f << "f" << i << "=" << val << ",";
 	}
 	obj.m_page =int(libwps::read16(m_input));
-	obj.m_origin=Vec2f(float(libwps::read16(m_input)/1440.),float(libwps::read16(m_input)/1440.));
+	int origin[2];
+	for (int i=0; i<2; ++i) origin[i] =int(libwps::read16(m_input));
+	obj.m_origin=Vec2f(float(origin[0]/1440.),float(origin[1]/1440.));
 	val =int(libwps::read16(m_input)); // can be big
 	if (val) f << "f2=" << val << ",";
 	obj.m_extra = f.str();
@@ -2563,7 +2566,7 @@ bool WPS4Text::readPLC
 				break;
 			default:
 				WPS_DEBUG_MSG(("WPS4Text:readPLC: unexpected PLC size\n"));
-			// fallthrough intended
+				WPS_FALLTHROUGH;
 			case 0:
 				plc.m_value = 0;
 			}
