@@ -22,6 +22,7 @@
 #include <stdlib.h>
 #include <string.h>
 
+#include <limits>
 #include <map>
 
 #include "libwps_internal.h"
@@ -250,12 +251,12 @@ std::ostream &operator<<(std::ostream &o, Object const &obj)
 struct Token
 {
 	//! constructor
-	Token() : m_type(WPSContentListener::None), m_textLength(-1), m_unknown(-1), m_text(""), m_error("") {}
+	Token() : m_type(WPSField::None), m_textLength(-1), m_unknown(-1), m_text(""), m_error("") {}
 	//! operator <<
 	friend std::ostream &operator<<(std::ostream &o, Token const &tok);
 
 	//! the field type
-	WPSContentListener::FieldType m_type;
+	WPSField::Type m_type;
 	//! the length of the text corresponding to the token
 	int m_textLength;
 	//! an unknown value
@@ -271,23 +272,29 @@ std::ostream &operator<<(std::ostream &o, Token const &tok)
 	o << std::dec;
 	switch (tok.m_type)
 	{
-	case WPSContentListener::PageNumber:
+	case WPSField::PageCount:
+		o << "field[pageCount],";
+		break;
+	case WPSField::PageNumber:
 		o << "field[page],";
 		break;
-	case WPSContentListener::Date:
+	case WPSField::PageNumberNext:
+		o << "field[nextpage],";
+		break;
+	case WPSField::Date:
 		o << "field[date],";
 		break;
-	case WPSContentListener::Time:
+	case WPSField::Time:
 		o << "field[time],";
 		break;
-	case WPSContentListener::Title:
+	case WPSField::Title:
 		o << "field[title],";
 		break;
-	case WPSContentListener::Link:
+	case WPSField::Link:
 		o << "field[link],";
 		break;
-	case WPSContentListener::Database:
-	case WPSContentListener::None:
+	case WPSField::Database:
+	case WPSField::None:
 	default:
 		o << "##field[unknown]" << ",";
 		break;
@@ -833,14 +840,18 @@ void WPS8Text::readText(WPSEntry const &entry)
 					switch (special.m_fieldType)
 					{
 					case WPS8TextStyle::FontData::F_PageNumber:
-						m_listener->insertField(WPSContentListener::PageNumber);
+						m_listener->insertField(WPSField(WPSField::PageNumber));
 						break;
 					case WPS8TextStyle::FontData::F_Date:
 					case WPS8TextStyle::FontData::F_Time:
 					{
 						std::string format = special.format();
 						if (format.length())
-							m_listener->insertDateTimeField(format.c_str());
+						{
+							WPSField field(WPSField::Date);
+							field.m_DTFormat=format;
+							m_listener->insertField(field);
+						}
 						else
 						{
 							WPS_DEBUG_MSG(("WPS8Text::readText: unknown date/time format for position : %lX\n", (unsigned long) pos));
@@ -1104,7 +1115,7 @@ bool WPS8Text::readString(RVNGInputStreamPtr input, long page_size,
 		if (!val) break;
 
 		long unicode = readUTF16LE(input, endString, val);
-		if (unicode != 0xfffd) WPSContentListener::appendUnicode(val, res);
+		if (unicode != 0xfffd) libwps::appendUnicode(val, res);
 	}
 	return true;
 }
@@ -1328,7 +1339,7 @@ void WPS8Text::createNotesCorrespondance()
 // plc: default, strs
 ////////////////////////////////////////////////////////////
 bool WPS8Text::defDataParser
-(long , long , int , WPS8Struct::FileData const &data, std::string &mess)
+(long, long, int, WPS8Struct::FileData const &data, std::string &mess)
 {
 	mess = "";
 	libwps::DebugStream f;
@@ -1384,7 +1395,7 @@ bool WPS8Text::textZonesDataParser
 			f << "###" << data << ",";
 			continue;
 		}
-		/* 1=Main, 2=Footnote, 3=Endnote, 5=Section/Table/Column/..,
+		/* 1=Main, 2=Footnote, 3=Endnote, 5=Contact/Section/Table/Column/..,
 		   6=Header, 7=Footer */
 		f << "id=" << dt.m_value << ",";
 		id = (int) dt.m_value;
@@ -1693,17 +1704,18 @@ bool WPS8Text::tokenEndDataParser(long endPage, std::vector<long> const &textPtr
 			case 1: // looks like the text size
 				tokn.m_textLength=int(dt.m_value);
 				break;
-			case 2: // -2=link, -5=page number, -6=?
+			case 2: // -2=link, -5=page number, -6=special char
 				switch (dt.m_value)
 				{
 				case -2:
-					tokn.m_type = WPSContentListener::Link;
+					tokn.m_type = WPSField::Link;
 					break;
 				case -5:
-					tokn.m_type = WPSContentListener::PageNumber;
+					tokn.m_type = WPSField::PageNumber;
+					break;
+				case -6:
+					f << "special[char],";
 					break;
-				// CHECKME: case -6-> strings = SPC, text character = 0xb7
-				// an insecable character or space ?
 				default:
 					WPS_DEBUG_MSG(("WPS8Text::tokenEndDataParser: unknown type=%d\n", int(dt.m_value)));
 					f << "###type=" << dt.m_value << ",";
@@ -1808,7 +1820,7 @@ bool WPS8Text::tokenEndDataParser(long endPage, std::vector<long> const &textPtr
 		}
 		f << "ptr=(";
 
-		if (pos+20+4*N > endPage)
+		if (N < 0 || N > std::numeric_limits<int>::max()/4 || 4*N > endPage-pos+20)
 		{
 			input->seek(pos, librevenge::RVNG_SEEK_SET);
 			return false;
